﻿#pragma TextEncoding	= "UTF-8"
#pragma rtGlobals 		= 3
#pragma IgorVersion		= 8.0
#pragma ModuleName		= SpectraGraphTools
#pragma version			= 2.0
#pragma DefaultTab		= {3,20,4}		// set default tab width in Igor Pro 9 and later

static Constant kVersion = 2.0
static StrConstant kVersionDate = "07/2023"

// --------------------- Project Updater header ----------------------
// If you're using Igor Pro 8 or later and have Project Updater
// installed, this package can check periodically for new releases.
// https://www.wavemetrics.com/project/Updater
static Constant kProjectID = 21562																// the project node on IgorExchange
static StrConstant ksShortTitle = "Spectra Graph Tools"											// the project short title on IgorExchange

//________________________________________________________________________________________________
//	Written by Stephan Thuermer - https://www.wavemetrics.com/user/chozo
//	Original code & idea for the right-click graph menus by Tony Withers - https://www.wavemetrics.com/user/tony
//
//	This is a collection of functions for modifying traces and graphs.
//	Contents:
//		- Basic Graph Style and Label Macros
//		- Quick Normalization :		Normalize to unit area or height / scale the traces in a graph
//		- SpectraWaterfall :		Generates a waterfall plot out of 2D data or a folder
//		- BuildTraceOffsetPanel :	A small panel to offset and color traces
//		- BuildGraphSizePanel :		A small panel to set the graph size and margins
//
// 2021-03-01 - ver. 1.00:	Initial public release.
// 2021-03-07 - ver. 1.10:	Quick normalization did not work properly when the data included NaNs.
//							Graph Size panel: Axes margins did not copy correctly from one graph to the other.
//							The Graph Size panel threw an error on Mac because of a wrong ModifyControl command.
//							Trace Offset panel: Added keyboard shortcuts for selecting and moving the cursor.
//							Trace Offset panel: It is now possible to move and scale selected traces with the cursor keys.
// 2021-03-11 - ver. 1.11:	A cursor set on an image triggered an error message for the TraceOffset panel.
// 2021-03-16 - ver. 1.20:	Added global constant for the multiplier step and made keyboard control more precise.
//							Added Offset Traces functionality to the graph and traces right-click menus.
// 2021-03-17 - ver. 1.21:	Now it is possible to choose whether offsets are set or added.
//							Fixed bug: Closing a graph did not properly update active window.
//							Now the total shift is also displayed for single traces.
// 2021-03-28 - ver. 1.30:	Graph Size panel: The last selected size unit setting is now saved in the current experiment.
//							Trace offset panel: Fixed a few error messages triggered by very old versions of the panel.
//							Trace offset panel: Trace scaling increment steps now get smaller and bigger depending on the
//							order of magnitude of the current scaling (helps to work with very big or small scaling factors).
//							Trace offset panel: Added a button to quickly set a new cursor onto the graph.
//							The currently selected trace name is now displayed in the panel title.
// 2021-05-14 - ver. 1.40:	Slightly adjusted the axis-label graph macros.
//							Graph size panel: Now switching the axes keeps the axis label visible.
//							Offset menu: Works now for 1-column 2D waves as well.
//							Graph Size panel: Hold shift while changing the font size to change the global axes font instead.
//							Graph Size panel: Added print checkbox to enable print-to-history function for most controls.
// 2021-08-14 - ver. 1.50:	Fixed bug: Commands got printed anyway when an old panel without the print checkbox was used.
//							Graph Size panel: Overhaul of the panel controls.
//							Graph Size panel: Added tick label margin controls.
//							Trace offset panel: Moved controls into tabs for better organization.
//							Trace offset panel: The Prev and Next cursor placement controls honor trace offsets now.
//							Trace offset panel: New cursors get set in the center of the current bottom-axis range.
// 2021-12-01 - ver. 1.60:	Added support for sub-graphs (also in layouts and panels) in all tools.
//							Trace offset panel: Fixed bug - cursors on shared image plots were not caught properly and led to errors.
//							Graph Size panel: Doubled the increment for size changes for faster adjustments.
//							Graph Size panel: Added controls for label margin.
//							Graph Size panel: Tick and label margins will be copied & pasted between graphs as well.
// 2022-03-23 - ver. 1.65:	Normalize to cursor now honors x-offsets of traces.
//							Added support for keyboard-shortcut (and other future) settings via external text file.
//							Added quick-scale tool to modify data directly from a graph.
//							Trace tool now shows color of currently selected trace.
// 2022-04-05 - ver. 1.70:	New right-click shortcut to add legend with live trace offset information.
//							Moved all graph related menu entries into Offset Spectra sub-menu.
//							Fixed bug: Quick-Scale panel did not react when trace together with cursor was removed.
//							Made the Quick-Scale panel x-wave aware.
// 2022-07-15 - ver. 1.80:	Quick-offset legend now ignores traces from ksTraceIgnoreList.
//							New keyboard shortcut for ScaleAllTraces() => hold ctrl to keep current offsets.
//							Added alphanumerical trace ordering.
//							Added new Normalize functions to graph's right-click menus and reorganized menu entries.
//							Live-offset legend: Support for trace instances and fixed some bugs.
//							Fixed Bug: Reverse sorting was not re-applying offsets correctly.
// 2022-10-19 - ver. 1.85:	Optimized code for ScaleAllTraces and added smoothing option to the right-click menu.
//							Added mean around cursor to ScaleAllTraces.
//							Added mouse cursor as a target for quick normalization.
//							Added normalization to the right-click menu of the graph marquee.
//							Added smoothing option for normalization.
//							Modifying traces now works for vertical graphs and non-standard axes.
// 2022-10-27 - ver. 1.90:	Trace offset panel: Filter now accepts lists and got a right-click menu to add traces to the list.
//							Trace offset panel: Fixed error when there is only one trace on the graph.
//							Graph Size panel: Added 'copy style to / paste style from clipboard' functionality.
//							Quick normalization can now be done to an arbitrary y value.
//							Fixed bug: Normalization to small values did not work properly.
//							All actions now work for vertical graphs and non-standard axes.
// 2023-07-25 - ver. 2.00:	Fixed bug: Quick-scaling failed sometimes when data had NaNs.
//							Fixed bug: Quick-scaling was not working for the zero-offset setting.
//							The Y Reset and X Reset buttons honor filter settings now.
//							Upgraded code to Igor 8.
//							Alphanumerical sorting now works for plots of 2D data.
//							Improved XY-plot support for scaling/offsetting and the Trace Tool.
//							Fixed various small inconsistencies and edge cases.
//							Fixed bug: Copy Style did not work properly with swapped axes and a fixed aspect ratio.
//							Graph Size and Trace Offset panels are now displayed next to the top graph and otherwise centered on the screen.
//							Overhauled the Trace Normalize menu to be less confusing.
//							Added 'Export Offsets' option to normalize menu.
//________________________________________________________________________________________________

static Constant kOffsetIncrStep = 0.005															// fraction of the total range to increase / decrease values each step
static Constant kMultiIncrStep  = 0.1															// by how much the multiplier is incremented (keyboard presses are 1/5 of this value)
static StrConstant ksTraceIgnoreList	= "w_base;MQP_*;*_tan;"									// list of traces to ignore, wildcards okay
static StrConstant ksGraphSizeUnitSave	= "V_SizePanelUnit"										// global variable name which saves the current unit setting of the Graph Size Panel
static StrConstant ksMenuOptionsSave	= "V_GTMenuOptions"										// global variable name for right-click menu options
static StrConstant ksMenuOptionsDefault	= "offsetMode:0;cursor:Mouse;cursorRange:3;normSmooth:0;waveExport:0;"	// default settings for right-click menus
static StrConstant kUsedCursorList		= "A;B;C;D;E;F;G;H;I;J;"
static StrConstant kSettingsFileName	= "Spectra Tools settings.dat"							// just for keyboard shortcuts for now.

//-------------------------------- functions for user code ---------------------------------------
static function TraceTool_FetchXShift(wave input, string &unit)									// extract predetermined X offset shift values from data
	variable xShift = NaN																		// the shift value which will be preset in the panel
	// ++++++++++++++++++++++ get the photon energy as shift value +++++++++++++++++++++++++++++++
	variable neg = 1
	string notes = note(input)
	string energy = StringByKey("Xray energy", notes, "=","\r")									// find the photon energy
	if (!strlen(energy))
		energy = StringByKey("photon energy ", notes, "=","\r")									// try a different key
	endif
	if (!strlen(energy))
		energy = StringByKey("Excitation Energy", notes, "=","\r")								// try yet a different key
		neg = -1																				// this data requires flipping
	endif
	unit = "eV"
	xShift = str2num(energy)*neg
	if (xShift == 0)
		xShift = NaN
	endif
	// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	return xShift 																				// return NaN for no change and return negative values for flipping the trace
end

static function TraceTool_FetchYShift(wave input, string &unit)									// extract predetermined Y offset shift values
	return NaN	 																				// return NaN for no change and return negative values for flipping the trace
end
//------------------------------------------------------------------------------------------------

//################################################################################################

Menu "Spectra Tools"
	"-"
	SubMenu "Normalize Traces"
		SpectraGraphTools#MenuEntry("norm_to1"),	/Q, ScaleAllTraces()						// hold shift for zero offset; hold alt to skip offset alignment
		SpectraGraphTools#MenuEntry("norm_toCur"),	/Q, ScaleAllTraces(csr="A")
		SpectraGraphTools#MenuEntry("norm_toArea"),	/Q, ScaleAllTraces(normMode=4)
		"-"
		"Smoothing Factor ...", /Q, SpectraGraphTools#SetSmoothOption()
		SpectraGraphTools#DisplayMenuOptions("zeroOffset"),	/Q, SpectraGraphTools#ToggleMenuOptions("zeroOffset")
		SpectraGraphTools#DisplayMenuOptions("keepOffset"),	/Q, SpectraGraphTools#ToggleMenuOptions("keepOffset")
		SpectraGraphTools#DisplayMenuOptions("waveExport"),	/Q, SpectraGraphTools#ToggleMenuOptions("waveExport")
	End
	"-"
	SpectraGraphTools#MenuEntry("waterfall"),	/Q, CreateBrowser; SpectraWaterfall()
	SpectraGraphTools#MenuEntry("quick_color"),	/Q, ColorAllTraces()
	SpectraGraphTools#MenuEntry("graph_GUI"),	/Q, BuildGraphSizePanel()
	SpectraGraphTools#MenuEntry("trace_GUI"),	/Q, BuildTraceOffsetPanel()
	SpectraGraphTools#MenuEntry("quick_scale"),	/Q, AttachQuickScalePanel()
End

Menu "GraphPopup"
	"-"
	SubMenu "Traces: Normalize"
		SubMenu "to One"
			SpectraGraphTools#DisplayNormMenu(), 			/Q, SpectraGraphTools#ScaleTracesWrapper("One")
		End
		SubMenu "to Global Maximum"
			SpectraGraphTools#DisplayNormMenu(), 			/Q, SpectraGraphTools#ScaleTracesWrapper("Global")
		End
			SpectraGraphTools#DisplayNormMenu(atCursor=1),	/Q, SpectraGraphTools#ScaleTracesWrapper("Cursor")
		"-"
		SubMenu "Select Cursor"
			SpectraGraphTools#DisplayCursorOptions(),		/Q, SpectraGraphTools#SetCursorOption()
		End
		"Smoothing Factor ...",								/Q, SpectraGraphTools#SetSmoothOption()
		SpectraGraphTools#DisplayMenuOptions("zeroOffset"),	/Q, SpectraGraphTools#ToggleMenuOptions("zeroOffset")
		SpectraGraphTools#DisplayMenuOptions("keepOffset"),	/Q, SpectraGraphTools#ToggleMenuOptions("keepOffset")
		SpectraGraphTools#DisplayMenuOptions("waveExport"),	/Q, SpectraGraphTools#ToggleMenuOptions("waveExport")
		"-"
		"Undo All Scaling",	/Q, ModifyGraph muloffset={0,0}
		"Undo All Offsets",	/Q, ModifyGraph offset={0,0}
	End
	Submenu "Traces: Offset and Sort"
		"y-Order Traces Alphanumerically",	/Q, SpectraGraphTools#OrderTracesAlphabetical()		// holding ctrl / command sorts in descending order
		"-"
		"Spread Traces Across y Range",		/Q, SpectraGraphTools#SpreadToFillRange()
		"Spread Traces: Equal Offset",		/Q, SpectraGraphTools#SpreadSpectra(0)
		"Spread Traces: Equal Gap",			/Q, SpectraGraphTools#SpreadSpectra(1)
		"-"
		"Undo All Scaling",	/Q, ModifyGraph muloffset={0,0}
		"Undo All Offsets",	/Q, ModifyGraph offset={0,0}
	End
	"-"
	"Add Annotation with Live Offset ...",	/Q, SpectraGraphTools#QuickOffsetLegendWrapper()
	"Left Axis: Autoscale Only Visible", 	/Q, SetAxis/A=2/E=0 left
	"Left Axis: Autoscale From Zero",	 	/Q, SetAxis/A=2/E=1 left
End

Menu "TracePopup"//, dynamic
	"-"
	SubMenu "Traces: Normalize"
		SubMenu "to One"
			SpectraGraphTools#DisplayNormMenu(), 			/Q, SpectraGraphTools#ScaleTracesWrapper("One")
		End
		SubMenu "to Global Maximum"
			SpectraGraphTools#DisplayNormMenu(), 			/Q, SpectraGraphTools#ScaleTracesWrapper("Global")
		End
			SpectraGraphTools#DisplayNormMenu(atCursor=1),	/Q, SpectraGraphTools#ScaleTracesWrapper("Cursor")
		"-"
		SubMenu "Select Cursor"
			SpectraGraphTools#DisplayCursorOptions(),		/Q, SpectraGraphTools#SetCursorOption()
		End
		"Smoothing Factor ...",								/Q, SpectraGraphTools#SetSmoothOption()
		SpectraGraphTools#DisplayMenuOptions("zeroOffset"),	/Q, SpectraGraphTools#ToggleMenuOptions("zeroOffset")
		SpectraGraphTools#DisplayMenuOptions("keepOffset"),	/Q, SpectraGraphTools#ToggleMenuOptions("keepOffset")
		SpectraGraphTools#DisplayMenuOptions("waveExport"),	/Q, SpectraGraphTools#ToggleMenuOptions("waveExport")
		"-"
		"Undo All Scaling",	/Q, ModifyGraph muloffset={0,0}
		"Undo All Offsets",	/Q, ModifyGraph offset={0,0}
	End
	Submenu "Traces: Offset and Sort"
		SpectraGraphTools#MenuWithinPlotrange("y-Offset: Align Traces Here"),	/Q, SpectraGraphTools#AlignTracesAtXPos(0)		// holding ctrl / command aligns to zero
		SpectraGraphTools#MenuWithinPlotrange("y-Scale: Match Traces Here"),	/Q, SpectraGraphTools#AlignTracesAtXPos(1)		// holding ctrl / command scales to one
		SpectraGraphTools#MenuWithinPlotrange("y-Order Traces Here"),			/Q, SpectraGraphTools#OrderTracesAtXPos()		// holding ctrl / command sorts in descending order
		"y-Order Traces Alphanumerically",										/Q, SpectraGraphTools#OrderTracesAlphabetical()	// holding ctrl / command sorts in descending order
		"-"
		"Spread Traces Across y Range", /Q, SpectraGraphTools#SpreadToFillRange()
		"Spread Traces: Equal Offset", 	/Q, SpectraGraphTools#SpreadSpectra(0)
		"Spread Traces: Equal Gap", 	/Q, SpectraGraphTools#SpreadSpectra(1)
		"-"
		"Undo All Scaling",	/Q, ModifyGraph muloffset={0,0}
		"Undo All Offsets",	/Q, ModifyGraph offset={0,0}
	End
	"-"
	"Add Annotation with Live Offset ...",	/Q, SpectraGraphTools#QuickOffsetLegendWrapper()
	"Left Axis: Autoscale Only Visible",	/Q, SetAxis/A=2/E=0 left
	"Left Axis: Autoscale From Zero",		/Q, SetAxis/A=2/E=1 left
End

Menu "GraphMarquee", dynamic
	SubMenu SpectraGraphTools#TraceMarqueeMenu("Traces: Normalize")
		SpectraGraphTools#DisplayMarqueeMenu(),		/Q, SpectraGraphTools#ScaleMarqueeWrapper()
		"-"
		"Smoothing Factor ...", /Q, SpectraGraphTools#SetSmoothOption()
		SpectraGraphTools#DisplayMenuOptions("zeroOffset"),	/Q, SpectraGraphTools#ToggleMenuOptions("zeroOffset")
		SpectraGraphTools#DisplayMenuOptions("keepOffset"),	/Q, SpectraGraphTools#ToggleMenuOptions("keepOffset")
		SpectraGraphTools#DisplayMenuOptions("waveExport"),	/Q, SpectraGraphTools#ToggleMenuOptions("waveExport")
		"-"
		"Undo All Scaling",	/Q, ModifyGraph muloffset={0,0}
		"Undo All Offsets",	/Q, ModifyGraph offset={0,0}
	End
End

//################################################################################################

static function/S MenuEntry(string which)
	string menuStr = "", searchStr = ""
	StrSwitch (which)
		case "norm_to1":
			menuStr = "to One"
			searchStr = "normalize traces to one"
		break
		case "norm_toCur":
			menuStr = "to One at Cursor A"
			searchStr = "normalize traces to cursor"
		break
		case "norm_toArea":
			menuStr = "to Area"
			searchStr = "normalize traces to area"
		break
		case "waterfall":
			menuStr = "Waterfall Plot of 2D Data or Folder"
			searchStr = "create waterfall plot"
		break
		case "quick_color":
			menuStr = "Quick Colorize Traces"
			searchStr = "quick colorize traces"
		break
		case "graph_GUI":
			menuStr = "Graph Size Panel ..."
			searchStr = "start graph size panel"
		break
		case "trace_GUI":
			menuStr = "Trace Offset Panel ..."
			searchStr = "start trace offset panel"
		break
		case "quick_scale":
			menuStr = "Attach Quick-Scale Panel ..."
			searchStr = "start quick-scale panel"
		break
		default:
			return ""
	EndSwitch
	
	int fileID; string read = ""
	Open/Z/R fileID as ParseFilePath(1, FunctionPath(""), ":", 1, 0)+kSettingsFileName
	if (!V_flag)
		do
			FReadLine fileID, read
		while (strlen(read) > 0 && !StringMatch(read, searchStr+"*"))
		read = cleanStr(read,"\t| |")
		read = StringByKey(cleanStr(searchStr,"\t| |"),read,"=","\r")
		Close fileID
	endif
	return menuStr+SelectString(strlen(read),"","/"+read)
end

//------------------------------------------------------------------------------------------------

static function/S TraceMarqueeMenu(string s_menu)
	return SelectString(ItemsInList(TraceNameList("",";",1)),"(","")+s_menu
end

static function ToggleMenuOptions(string which)
	string settings = StrVarOrDefault("root:"+ksMenuOptionsSave,ksMenuOptionsDefault)
	StrSwitch(which)
		case "waveExport":
			settings = ReplaceNumberByKey("waveExport" ,settings ,(NumberByKey("waveExport"	,settings) == 1 ? 0 : 1))
		break
		default:
			int mode = WhichListItem(which, "zeroOffset;keepOffset;")+1
			if (mode >= 1)
				settings = ReplaceNumberByKey("offsetMode" ,settings ,(NumberByKey("offsetMode"	,settings) == mode ? 0 : mode))
			endif
		break
	EndSwitch
	string/G root:$ksMenuOptionsSave = settings
	BuildMenu "GraphPopup"
	BuildMenu "TracePopup"
	return 0
end

static function/S DisplayCursorOptions()
	string settings = StrVarOrDefault("root:"+ksMenuOptionsSave,ksMenuOptionsDefault)
	string csr = StringByKey("cursor" ,settings); int mouse = CmpStr(csr,"Mouse")==0
	string cursorList = ReplaceString(csr,kUsedCursorList,"!"+num2char(18)+csr)
	return SelectString(mouse,"","!"+num2char(18))+"Mouse;"+cursorList
end

static function/S DisplayMenuOptions(string which)
	string settings = StrVarOrDefault("root:"+ksMenuOptionsSave,ksMenuOptionsDefault)
	StrSwitch(which)
		case "zeroOffset":
			return SelectString(NumberByKey("offsetMode" ,settings) == 1,"","!"+num2char(18)) + "Keep y-Offset at Zero"
		break
		case "keepOffset":
			return SelectString(NumberByKey("offsetMode" ,settings) == 2,"","!"+num2char(18)) + "Keep Current y-Offset"
		break
		case "waveExport":
			return SelectString(NumberByKey("waveExport" ,settings) == 1,"","!"+num2char(18)) + "Export Offsets to Wave"
		break
	EndSwitch
end

static function SetCursorOption()
	string settings = StrVarOrDefault("root:"+ksMenuOptionsSave,ksMenuOptionsDefault)
	GetLastUserMenuInfo
	string/G root:$ksMenuOptionsSave = ReplaceStringByKey("cursor",settings,S_Value)
	BuildMenu "GraphPopup"
	BuildMenu "TracePopup"
	return 0
end

static function SetSmoothOption()
	string settings = StrVarOrDefault("root:"+ksMenuOptionsSave,ksMenuOptionsDefault)
	variable smth = NumberByKey("normSmooth",settings)
	Prompt smth, "Set Smoothing Factor for Normalization (0 = off):"
	DoPrompt "Smoothing for Trace Normalization", smth
	if (!V_Flag)
		string/G root:$ksMenuOptionsSave = ReplaceNumberByKey("normSmooth",settings,smth)
	endif
	return 0
end

static function/S DisplayNormMenu([int atCursor])
	string settings = StrVarOrDefault("root:"+ksMenuOptionsSave,ksMenuOptionsDefault)
	string csr = StringByKey("cursor" ,settings); int mouse = CmpStr(csr,"Mouse")==0
	csr = SelectString(mouse,"Cursor ","") + StringByKey("cursor" ,settings) + SelectString(mouse,""," Cursor")
	if (atCursor)
		return "to "+csr+" Position"
	else
		return "Maxima;Area;at "+csr+" x Position;Maxima around "+csr+";Average around "+csr+";"
	endif
end

//________________________________________________________________________________________________
//
//										Standard Graph Macros
//________________________________________________________________________________________________

Proc BoxStyle() : GraphStyle
	PauseUpdate; Silent 1
	GetAxis/Q right
	if (V_flag == 1)
		ModifyGraph/Z mirror(left)=2
	endif
	GetAxis/Q top
	if (V_flag == 1)
		ModifyGraph/Z mirror(bottom)=2
	endif
	ModifyGraph noLabel=0
	ModifyGraph standoff=0
	ModifyGraph ZisZ=1
	ModifyGraph tick=0
	ModifyGraph notation(left)=1
	ModifyGraph axisOnTop=1
	ModifyGraph minor=1
	if ((GetKeyState(0) & 2^0) != 0)															// ctrl / command = paper style (no left axis notation)
		ModifyGraph tick(left)=3
		ModifyGraph noLabel(left)=1
	endif
	ResumeUpdate
EndMacro

Proc XAS_Label() : GraphStyle
	Label/Z left "\u#2Intensity [arb. u.] \\E"
	Label/Z bottom "Photon energy [\U]"
EndMacro

Proc XES_Label() : GraphStyle
	Label/Z left "\u#2Intensity [arb. u.] \\E"
	Label/Z bottom "Emission energy [\U]"
EndMacro

Proc TOF_Label() : GraphStyle
	Label/Z left "\u#2Intensity [arb. u.] \\E"
	Label/Z bottom "Flight time [\U]"
EndMacro

Proc PKE_Label() : GraphStyle
	Label/Z left "\u#2Intensity [arb. u.] \\E"
	Label/Z bottom "\u#2"+SelectString((GetKeyState(0) & 2^0) != 0, "Kinetic", "Binding")+" energy [eV]"	// BE label with ctrl / command
EndMacro

//________________________________________________________________________________________________
//	Change the Color of up to 13 Traces with Preset Colors - Sorted to give good contrast for even a few traces
//________________________________________________________________________________________________

function ColorAllTraces()																		// colors traces with 12 different colors (hard-coded)
	string traces, gName = getTopGraph()
	variable trNum = FetchAllTraces(gName,traces)
	if (trNum == 0)
		return 0
	endif
	Make/FREE colors = {{65280,0,0}, {65280,43520,0}, {0,65280,0}, {0,52224,0}, {0,26214,1329}, {0,65280,65280}, {0,43520,65280}, {0,15872,65280}, {44253,29492,58982}, {65535,16385,55749}, {26411,1,52428}, {26112,26112,26112}, {0,0,0}}
	variable trace, color, cCount = DimSize(colors,1)
	variable leap = trNum < cCount/2 ? trunc((cCount-1)/(trNum-1)) : 1							// introduce a leap value to choose more distant colors for only a few traces
	for (trace = 0, color = 0; trace < trNum; trace += 1, color += leap)
		color = color == cCount ? 0 : color														// loop colors around
		ModifyGraph/W=$gName rgb($StringFromList(trace,traces))=(colors[0][color],colors[1][color],colors[2][color])	// set new color offset
	endfor
end

//________________________________________________________________________________________________
//
//						Add Legend with Trace Offset and Scaling Infos
//________________________________________________________________________________________________

static function InsertIntoMainProc(string newCode)
	string currScrap = GetScrapText()															// copy current scrap text
	GetWindow Procedure hide; int wasHidden = V_Value != 0
	DisplayProcedure/W=Procedure
	DoIgorMenu "Edit" "Select All"
	DoIgorMenu "Edit" "Copy"
	PutScrapText GetScrapText()+newCode															// modify procedure code
	DoIgorMenu "Edit" "Paste"
	PutScrapText currScrap			 															// put previous scrap text back
	Execute/P/Q/Z "COMPILEPROCEDURES "															// recompile all
	if (wasHidden)
		HideProcedures																			// hide all procedure windows
	endif
end
//------------------------------------------------------------------------------------------------
static function/S OffsetFuncCode()																// inserts function for a live display of trace offset
	string newCode = "\r\rfunction/S GT_liveOff(string gName, string trace, variable which)\r"
	newCode += "\tif (which < 0 || which > 3)\t\t// which modes: [0] = x offset, [1] = y offset, [2] = x multiplier, [3] = y multiplier\r\t\treturn \"\"\r\tendif\r"
//	newCode += "\tvariable instance = str2num(StringFromList(1,trace,\"#\")); instance = numtype(instance) == 0 ? instance : 0\r"
	newCode += "\tvariable mode = which > 1, xval, yval, roundTo = NumVarOrDefault(\"root:GT_liveOffRoundValues\",5)\r"
	newCode += "\tsscanf StringByKey(SelectString(mode,\"offset(x)\",\"muloffset(x)\"), TraceInfo(gName,trace,0), \"=\"), \"{%f,%f}\", xval, yval\r"
	newCode += "\tstring result;\tsprintf result, \"%.*g\", roundTo, (which-2*mode ? yval+(yval==0)*mode : xval)\r\treturn result\rend"
	return newCode
end

//################################################################################################

static function QuickOffsetLegendWrapper()
	variable mode = 1, add = 1, pos = 1, rnd = NumVarOrDefault("root:GT_liveOffRoundValues",5)+1
	string hlpStr = "This tool adds a function GT_liveOff(graphName, traceName, mode) to the current experiment to display live offset information of traces."
	Prompt mode,"1st Information:",popup,"x Offsets;y Offsets;x Multiplier;y Multiplier;"
	Prompt add, "2nd Information:",popup,"None;x Offsets;y Offsets;x Multiplier;y Multiplier;"
	Prompt pos, "Legend Position:",popup,"Default;LB: left bottom;MB: mid bottom;RB: right bottom;LC: left center;MC: mid center;RC: right center;LT: left top;MT: mid top;RT: right top;"
	Prompt rnd, "No. of Significant Digits (Rounding):",popup,"0;1;2;3;4;5;6;7;8;9;10;"
	DoPrompt/HELP=hlpStr "Which information should be displayed in the legend of "+getTopMainGraph()+"?",mode, add, rnd, pos
	if (!V_Flag)
		QuickOffsetLegend(mode-1, addinfo = add-2, anchor = pos-1, digits = rnd-1)	
	endif
	return 0
end

//------------------------------------------------------------------------------------------------

function QuickOffsetLegend(int which, [int addinfo, int anchor, int digits])					// which modes: [0] = x offset, [1] = y offset, [2] = x multiplier, [3] = y multiplier
	string gName = getTopMainGraph()
	if (!strlen(gName))
		return -1
	endif
	
	int anchorNo= ParamIsDefault(anchor) || anchor < 0   || anchor >  9 ?  7 : anchor			// default is LT anchor
	int doAdd	= ParamIsDefault(addinfo)|| addinfo < -1 || addinfo>  3 ? -1 : addinfo			// default is no info
	int roundTo	= ParamIsDefault(digits) || digits < 0   || digits > 16 ?  5 : digits			// default is 5 significant digits
	if (!strlen(FunctionInfo("GT_liveOff","Procedure")))										// first inject offset macro into experiment
		InsertIntoMainProc(OffsetFuncCode())
	endif
	
	variable/G root:GT_liveOffRoundValues = roundTo
	string modeList = "Δx = ;Δy = ;x * ;y * ;"
	string anchorID = StringFromList(anchorNo,"none;LB;MB;RB;LC;MC;RC;LT;MT;RT;")
	string traces = "", legendText = ""
	
	int i, items = FetchAllTraces(gName,traces)
	for (i = 0; i < items; i++)
		string currTrace = StringFromList(i, traces)
		string cleanName = cleanStr(currTrace, "'|_prH|_prV|_prL|")								// cleanup endings from image profiles
		legendText += "\\s("+currTrace+") "+cleanName
		legendText += " ("+StringFromList(which, modelist)+"\\{GT_liveOff(\""+gName+"\",\""+currTrace+"\", "+num2str(which)+")}"
	if (doAdd > -1)
		legendText += ", "+StringFromList(doAdd, modelist)+"\\{GT_liveOff(\""+gName+"\",\""+currTrace+"\", "+num2str(doAdd)+")}"
	endif
		legendText += ")\r"
	endfor
	legendText = RemoveEnding(legendText,"\r")
	
	if (CmpStr (anchorID, "none") == 0)
		Legend/W=$gName/C/N=TraceOffsets/J/F=0/B=1 legendText
	else
		Legend/W=$gName/A=$anchorID/C/N=TraceOffsets/J/F=0/B=1 legendText
	endif
	return 0
end

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static function/S cleanStr(string str, string list)
	int i
	for (i = 0; i < ItemsInList(list,"|"); i++)
		str = ReplaceString(StringFromList(i,list,"|"), str, "")
	endfor
	return str
end

//________________________________________________________________________________________________
//
//							Scaling Normalization of Graph Traces
//________________________________________________________________________________________________

static function ScaleTracesWrapper(string type)													// call ScaleAllTraces from right-click menu
	GetLastUserMenuInfo
	string gName = S_graphName, trace = S_traceName, selText = S_Value;	int selVal = V_Value
	if (V_flag != 0)
		return 0
	endif
	
	int toArea = StringMatch(selText,"*Area*")
	int doCsr = limit(selVal-2, 0, inf)															// cursor related entries start from the third one
	int rMode = limit(selVal-3, 0, inf)
	int nMode = 0
	nMode = !CmpStr(type,"One") ?	 0+3*toArea : nMode
	nMode = !CmpStr(type,"Global") ? 1+3*toArea : nMode
	nMode = !CmpStr(type,"Cursor") ? 2 : nMode
	
	string opt = StrVarOrDefault("root:"+ksMenuOptionsSave,ksMenuOptionsDefault)				// load options
	string csr = SelectString(doCsr>0,"None",StringByKey("cursor" ,opt))
	variable noOff	 = (NumberByKey("offsetMode",opt) == 1) || ((GetKeyState(0)&2^2) != 0)		// don't shift background if shift is pressed
	variable keepOff = (NumberByKey("offsetMode",opt) == 2) || ((GetKeyState(0)&2^0) != 0)		// don't touch offset if ctrl is pressed
	variable offMode = keepOff ? 2 : (noOff ? 1 : 0)
	variable xRange  = NumberByKey("cursorRange",opt)/100										// x range around cursor to look for maximum in percent
	variable doSmth	 = NumberByKey("normSmooth", opt)											// smoothing factor
	variable wExport = NumberByKey("waveExport", opt);	wExport = numtype(wExport) != 0 ? 0 : wExport
	
	variable xpos = NaN, ypos = NaN
	if (CmpStr(csr,"Mouse") == 0)																// get mouse position
		string xAxis = "", yAxis = "",traces = ""
		if (strlen(trace))
			xAxis = StringByKey("XAXIS",TraceInfo(gName,trace,0))
			yAxis = StringByKey("YAXIS",TraceInfo(gName,trace,0))
		else
			if (FetchAllTraces(gName,traces) == 0)												// no traces found
				return -1
			endif
			xAxis = GetMajorAxisFromTraceList(gName, traces, "X")
			yAxis = GetMajorAxisFromTraceList(gName, traces, "Y")
		endif
		variable axType = FindListItem(StringByKey("AXTYPE",AxisInfo(gName,xAxis)), "top;bottom;") > -1		// which type of axis do we work with: left / right or top / bottom?
		xpos = axType ? AxisValFromPixel(gName, xAxis, V_mouseX) : AxisValFromPixel(gName, xAxis, V_mouseY)
		ypos = axType ? AxisValFromPixel(gName, yAxis, V_mouseY) : AxisValFromPixel(gName, yAxis, V_mouseX)
	endif
	
	ScaleAllTraces(name=gName,smth=doSmth,xpos=xpos,ypos=ypos,normMode=nMode,offMode=offMode,rangeMode=rMode,relRange=xRange,csr=csr,export=wExport)
	return 0
end

//------------------------------------------------------------------------------------------------

static StrConstant kMarqueeMenu = "Maxima to One;Maxima to Global Maximum;Average to One;Average to Global Average;Area to One;Area to Global Maximum;"

static function/S DisplayMarqueeMenu()
	return kMarqueeMenu
end

static function ScaleMarqueeWrapper()															// call ScaleAllTraces from graph marquee
	GetLastUserMenuInfo
	string gName = S_graphName, traces = "", selText = S_Value;	int selVal = V_Value
	
	int toArea = StringMatch(selText,"*Area*")
	int rMode = StringMatch(selText,"*Average*") ? 2 : 1
	int nMode = 0
	nMode = StringMatch(selText,"*One*") ?    0+3*toArea : nMode
	nMode = StringMatch(selText,"*Global*") ? 1+3*toArea : nMode
	
	if (FetchAllTraces(gName,traces) == 0)
		return -1
	endif
	GetMarquee/K/W=$gName/Z $GetMajorAxisFromTraceList(gName, traces, "X")
	if (!V_flag)
		return -1
	endif
	variable xpos, xRange
	if (V_top == 0 && V_bottom == 0)															// horizontal axis
		xpos   = (V_left+V_right)/2
		xRange = abs(V_right-V_left)
	else																						// vertical axis
		xpos   = (V_top+V_bottom)/2
		xRange = abs(V_bottom-V_top)
	endif
	string opt = StrVarOrDefault("root:"+ksMenuOptionsSave,ksMenuOptionsDefault)
	variable noOff	 = (NumberByKey("offsetMode",opt) == 1) || ((GetKeyState(0)&2^2) != 0)		// don't shift background if shift is pressed
	variable keepOff = (NumberByKey("offsetMode",opt) == 2) || ((GetKeyState(0)&2^0) != 0)		// don't touch offset if ctrl is pressed
	variable offMode = keepOff ? 2 : (noOff ? 1 : 0)
	variable doSmth  = NumberByKey("normSmooth",opt)											// smoothing factor
	variable wExport = NumberByKey("waveExport", opt);	wExport = numtype(wExport) != 0 ? 0 : wExport
	ScaleAllTraces(name=gName,smth=doSmth,xpos=xpos,normMode=nMode,offMode=offMode,rangeMode=rMode,xRange=xRange,export=wExport)
	return 0
end

//------------------------------------------------------------------------------------------------

function ScaleTo1()																				// backwards compatibility
	ScaleAllTraces()
end

//------------------------------------------------------------------------------------------------

function ScaleAllTraces([string name, variable smth, variable xpos, variable ypos, int normMode, int offMode, int rangeMode, variable xRange, variable relRange, string csr, int export])
	// these modes alter the behavior of normalization:
	int nMode = ParamIsDefault(normMode)  ? 0 : normMode		// how to normalize:		[0] = to one,  [1] = to global max,	[2] = to given y-value, [3] = area to one, [4] = area to global area	
	int oMode = ParamIsDefault(offMode)   ? 0 : offMode			// how to handle offsets:	[0] = to zero, [1] = no offset,		[2] = keep offset
	int rMode = ParamIsDefault(rangeMode) ? 0 : rangeMode		// use range around xPos:	[0] = off,     [1] = find maximum,	[2] = calculate average
	
//	Print "normalize:", nMode, "offset:", oMode, "range:", rMode, "position X/Y:", xpos, ypos	// debug
	
	string gName = SelectString(ParamIsDefault(name), name, getTopGraph())
	if (WinType(gName) != 1)
		Print "No graph with name "+name+"."
		return 0
	endif
	string traces = TraceNameList(gName,";",1)
	int i, items  = ItemsInList(Traces)
	if (items == 0)
		return 0
	endif
	
	int useGlobScale = nMode == 1 || nMode == 4													// scale to global maximum (highest intensity of all traces)
	int useYposScale = nMode == 2																// use given y-value
	int useAreaScale = nMode == 3 || nMode == 4													// scale to trace area
	int doExport = ParamIsDefault(export) ? 0 : export
	variable xPosVal = ParamIsDefault(xpos) ? NaN : xpos										// scale at specific x position
	variable yPosVal = ParamIsDefault(ypos) ? NaN : ypos
	variable relatXRange = ParamIsDefault(relRange) ? 0.03 : limit(relRange,0.001,1)			// default range is +-3 % of the total wave range
	variable xRangeWidth = ParamIsDefault(xRange) ? NaN : xRange
	
	variable xOff = 0, yOff = 0, xScale = 0, yScale = 0
	if (!ParamIsDefault(csr) && WhichListItem(csr, kUsedCursorList,";", 0, 0) > -1)
		string info = CsrInfo($csr,gName)
		if (!strlen(info))
			Abort "Cursor "+csr+" is not set."
		endif
		getAllTraceOffsets(gName, StringByKey("TNAME",info), xOff, yOff, xScale, yScale)
		xPosVal = hcsr($csr, gName)*xScale + xOff												// add trace offset to the cursor position
		yPosVal = vcsr($csr, gName)*yScale + yOff
	endif
	
	string curTrace, columnStr
	Make/D/Free/N=(items) minVals = NaN, maxVals = NaN, maxArea = NaN, minArea = NaN
	for (i = 0; i < items; i++)
		curTrace = StringFromList(i,traces)
		wave  w = prepareWaveFromTrace(gName, curTrace)
		wave/Z w_x = XWaveRefFromTrace(gName, curTrace)
		
		int doXwave = WaveExists(w_x)
		if( !ParamIsDefault(smth) && smth > 0)													// optional smoothing before scaling
			Smooth limit(smth,1,2^15-1), w
		endif
		
		minVals[i] = wavemin(w)
		maxVals[i] = wavemax(w)
		if (doXwave)
			SetScale/P x, 0, 1, w																// make sure scaling is removed
			maxArea[i] = areaXY(w,w_x)
			minArea[i] = wavemin(w)*abs(wavemax(w_x)-wavemin(w_x))
		else
			maxArea[i] = area(w)
			minArea[i] = wavemin(w)*DimDelta(w,0)*(DimSize(w,0)-1)								// area baseline
		endif
		
		if (numtype(xPosVal) == 0)
			getAllTraceOffsets(gName, curTrace, xOff, yOff, xScale, yScale)						// adjust for each trace's offsets
			variable scaledXpos = (xPosVal - xOff)/xScale
			variable xRng = numtype(xRangeWidth) == 0? xRangeWidth/2 : abs(DimDelta(w,0)*DimSize(w,0)*relatXRange)	// range in percent of the total wave range
			
			int pMin, pMax
			if(doXwave)
				FindLevel/P/Q w_x, scaledXpos-xRng;	pMin = !v_flag ? V_LevelX : 0
				FindLevel/P/Q w_x, scaledXpos+xRng;	pMax = !v_flag ? V_LevelX : DimSize(w,0)-1
				WaveStats/Q/R=[pMin, pMax] w
				Switch (rMode)
					case 1:
						maxVals[i] = V_max
					break
					case 2:
						maxVals[i] = V_avg
					break
					default:
						FindLevel/P/Q w_x, scaledXpos
						maxVals[i] = w[limit(V_LevelX,0,DimSize(w,0)-1)]
					break
				EndSwitch
				maxArea[i] = areaXY(w,w_x,w_x[pMin],w_x[pMax])
				minArea[i] = wavemin(w)*abs(w_x[pMax]-w_x[pMin])
			else
				WaveStats/Q/R=(scaledXpos-xRng, scaledXpos+xRng) w
				Switch (rMode)
					case 1:
						maxVals[i] = V_max
					break
					case 2:
						maxVals[i] = V_avg
					break
					default:
						maxVals[i] = w[limit(x2pnt(w,scaledXpos),0,DimSize(w,0)-1)]
					break
				EndSwitch
				maxArea[i] = area(w,scaledXpos-xRng, scaledXpos+xRng)
				minArea[i] = wavemin(w)*xRng*2
			endif
		endif
	endfor
	
	MatrixOP/O maxVals = replaceNaNs(maxVals,1)
	MatrixOP/O minVals = replaceNaNs(minVals,0)
	MatrixOP/O maxArea = replaceNaNs(maxArea,1)
	MatrixOP/O minArea = replaceNaNs(minArea,1)
	
	WaveStats/Q maxVals;	variable targetMax  = useGlobScale ? V_max-minVals[V_maxLoc]*(!oMode) : 1
	WaveStats/Q maxArea;	variable targetArea = useGlobScale ? V_max-minArea[V_maxLoc]*(!oMode) : 1
	targetMax  = useYposScale ? yPosVal : targetMax
	targetMax  = numtype(targetMax)  != 0 || targetMax  == 0 ? 1 : targetMax
	targetArea = numtype(targetArea) != 0 || targetArea == 0 ? 1 : targetArea
	
	Make/Free/D/N=(items,2) outWave=NaN
	SetDimLabel 1, 0, $("y Offset"), outWave
	SetDimLabel 1, 1, $("y Multiplier"), outWave
	
	variable scale
	for (i = 0; i < items; i++)
		string currTrace = StringFromList(i,traces)
		variable currMin  = !oMode ? minVals[i] : 0
		variable baseArea = !oMode ? minArea[i] : 0
		if (useAreaScale)
			scale = (maxArea[i]-baseArea)/targetArea 											// scale to match global area or to unit area
		else
			scale = (maxVals[i]-currMin)/targetMax 												// scale to match global maximum or to one
		endif
		scale = scale == 0 ? 1 : scale
		
		string cleanName = cleanStr(currTrace,"\"|'|,|;|")[0,255]
		SetDimLabel 0, i, $cleanName,  outWave
		
		Switch (oMode)
			case 0:		// adjust to Y=0
				ModifyGraph/W=$gName offset($currTrace)={,-currMin/scale}
				outWave[i][0] = -currMin/scale
			break
			case 1:		// offset = 0
				ModifyGraph/W=$gName offset($currTrace)={,0}
				outWave[i][0] = 0
			break
			case 2:		// keep offset => do nothing
				outWave[i][0] = getTraceOffset(gName, currTrace, 1)
			break
		EndSwitch
		
		outWave[i][1] = 1/scale
		scale = scale == 1 ? 0 : 1/scale														// no scaling for maximal trace
		ModifyGraph /W=$gName muloffset($currTrace)={,scale}
	endfor
	
	if (doExport)
		Note outWave, "Scaling and offsets for traces in graph '"+gName+"'\rOutput done on " + Secs2Date(DateTime,-2) + ", " + time()
		Duplicate/O outWave, $("Offsets_"+gName)
	endif
	return 0
end

//________________________________________________________________________________________________
//
//					Creates Waterfall Plot from a 2D wave or a Folder of Waves
//________________________________________________________________________________________________

function SpectraWaterfall()
	wave/Z inwave = $GetBrowserSelection(0)														// get the first selected wave
	if (!WaveExists(inwave))
		Print "Nothing selected"
		return -1
	endif
	DFREF saveDFR = GetDataFolderDFR()
	
	string list	= ""
	int is1D = WaveDims(inwave)==1
	if (is1D)
		SetDataFolder GetWavesDataFolderDFR(inwave)
		list = SortList(WaveList("!Int_*",";","DIMS:1"), ";", 16)
	endif
	
	string wTitle = SelectString(is1D,NameOfWave(inwave),GetWavesDataFolder(inwave,0))			// name of 2D wave or folder
	Display/W=(230,50,650,450)/K=1/N=$UniqueName("WaterfallGraph", 6, 0) as "Waterfall of "+wTitle
	int i, items = is1D ? ItemsInList(list) : DimSize(inwave,1)
	for (i=items-1; i>-1; i--)																	// go backwards (in reverse order, so that the later traces are BEHIND the earlier ones)
		if (is1D)
			AppendToGraph/Q $StringFromList(i,list)
		else
			AppendToGraph/Q inwave[][i]
		endif
	endfor
	SetDataFolder saveDFR
	
	Execute/Q "BoxStyle()"
	ColorSpectrumFromTable(0, "Rainbow")
	
	list = TraceNameList("",";",1)
	variable initOffset	= WaveMax(inwave) * 0.05												// the initial distance will be set in percent from the maximal value of the selected wave
	for (i = 1; i < items; i++)
		ModifyGraph offset($StringFromList(items-1 - i,list))={,i*initOffset}
	endfor
	return 0
end

//________________________________________________________________________________________________
//
//							Right-Click Offset Menus for Graph Traces
//				Adapted from the StackTraces procedures v1.20 by Tony Withers
//________________________________________________________________________________________________

static function/S MenuWithinPlotrange(string s_menu)											// use to create a menu item in a popup when you need to know where the user clicks to invoke the popup
	if (WinType("") != 1)																		// don't create variables if Igor is just rebuilding the menu
		return ""
	endif
	GetMouse/W=kwTopWin
	if (V_left<0 || V_top<0)
		return ""
	endif
	return s_menu
end

//################################################################################################

static function getYfromWave(string gName, string trace, variable xPos, variable &yVal, variable &yMin)
	wave w = prepareWaveFromTrace(gName, trace)
	wave/Z w_x = XWaveRefFromTrace(gName, trace)
	
	yMin = WaveMin(w);	yVal = NaN
	if(WaveExists(w_x))
		FindLevel/P/Q w_x, xPos
		if (!v_flag && v_levelX > 0 && v_levelX < DimSize(w,0)-1)								// make sure to stay within range
			yVal = w[round(v_levelX)]
		endif
	else
		if (x2pnt(w, xPos) > 0 && x2pnt(w, xPos) < DimSize(w,0)-1)
			yVal = w(xPos)
		endif
	endif
	return numtype(yVal)																		// error return
end

//################################################################################################

static function getMinMaxfromWave(string gName, string trace, variable &yMin, variable &yMax)	// figure out plotted x-range for trace and finds wave's max and min values in this range
	wave w = prepareWaveFromTrace(gName, trace)
	wave/Z w_x = XWaveRefFromTrace(gName, trace)
	
	variable xOff = 0, yOff = 0, xScale = 0, yScale = 0											// x-offset and y-scale aware
	getAllTraceOffsets(gName, trace, xOff, yOff, xScale, yScale)
	GetAxis/W=$gName/Q $StringByKey("XAXIS",TraceInfo(gName,trace,0))
	if(WaveExists(w_x))
		FindLevel /Q w_x, v_min;	v_min = !v_flag ? V_LevelX : 0								// no /P flag here - assumes same x-scaling for w and w_x
		FindLevel /Q w_x, v_max;	v_max = !v_flag ? V_LevelX : DimSize(w,0)-1
	endif
	variable fr = (v_min - xOff)/xScale
	variable to = (v_max - xOff)/xScale
	yMax = WaveMax(w, fr, to)*yScale
	yMin = WaveMin(w, fr, to)*yScale
	return 0
end

//################################################################################################

static function/wave prepareWaveFromTrace(string gName, string trace)							// extracts the correct column and removes NaNs
	wave w = TraceNameToWaveRef(gName, trace)
	Duplicate/FREE w temp
	if (WaveDims(w) == 2 && DimSize(w,1) > 1)													// handle 2D waterfall plots => extract column
		string columnStr
		SplitString/E=("\]\[(.*)\]") StringByKey("YRANGE",TraceInfo(gName,trace,0)), columnStr	// find the displayed column from TraceInfo
		variable col = str2num(columnStr)
		if (numtype(col) == 0)
			Duplicate/R=[][col]/FREE w temp
		endif
	endif
	Redimension/N=(-1,1) temp																	// needs 2D data for MatrixFilter
	MatrixFilter NanZapMedian temp																// make sure there are no NaNs
	return temp
end

//################################################################################################

static function AlignTracesAtXPos(int which)													// use trace popup to y-align traces at clicked x position by setting offsets (which = 0) or multipliers (which = 1)
	GetLastUserMenuInfo																			// figure out graph and trace names
	string gName = S_graphName, trace = S_traceName, traces = ""
	int trNum = FetchAllTraces(gName,traces)
	if (!trNum)
		return -1
	endif
	
	string xAxis = StringByKey("XAXIS",TraceInfo(gName, trace, 0)), yAxis = StringByKey("YAXIS",TraceInfo(gName, trace, 0))
	variable axType = FindListItem(StringByKey("AXTYPE",AxisInfo(gName,xAxis)), "top;bottom;") > -1			// which type of axis do we work with: left / right or top / bottom?
	variable xClick = axType ? AxisValFromPixel(gName, xAxis , V_mouseX) : AxisValFromPixel(gName, xAxis , V_mouseY)
	variable yClick = axType ? AxisValFromPixel(gName, yAxis , V_mouseY) : AxisValFromPixel(gName, yAxis , V_mouseX)
	variable doNorm = ((GetKeyState(0) & 2^0)!=0)												// holding ctrl / command aligns to zero or scales to one
	yClick = doNorm && which==0 ? 0 : yClick													// shift to zero?
	
	int i
	variable modVal
	for(i = 0; i < trNum; i++)
		trace = StringFromList(i,traces)
		variable xOff = 0, yOff = 0, xScale = 0, yScale = 0, ywVal, ywMin
		getAllTraceOffsets(gName, trace, xOff, yOff, xScale, yScale)
		int err = getYfromWave(gName, trace, (xClick-xOff)/xScale, ywVal, ywMin)				// extract wave min and Y value at X position
		if (err)
			continue
		endif
		if (which)
			modVal  = doNorm ? 1 : (yClick-(yOff+ywMin)/yScale)									// scale to one?
			modVal /= (ywVal-ywMin/yScale)
			ModifyGraph/W=$gName muloffset($trace)={,modVal}									// scale to match position
			if (doNorm)																			// if normalized scale offsets down as well (because the scale difference can be huge)
				ModifyGraph/W=$gName offset($trace)={,-ywMin*modVal}
			endif
		else
			modVal = yClick-ywVal*yScale
			ModifyGraph/W=$gName offset($trace)={,modVal}										// shift to align position
		endif
	endfor
	return 0
end

//################################################################################################

static function OrderTracesAtXPos()																// reorder traces based on (possibly offset) value at x=xpos
	GetLastUserMenuInfo																			// figure out graph and trace names
	string gName = S_graphName, trace = S_traceName, traces = ""
	int trNum = FetchAllTraces(gName,traces)
	if (!trNum)
		return -1
	endif
	
	string xAxis = StringByKey("XAXIS",TraceInfo(gName, trace, 0))
	variable axType = FindListItem(StringByKey("AXTYPE",AxisInfo(gName,xAxis)), "top;bottom;") > -1			// which type of axis do we work with: left / right or top / bottom?
	variable xClick = axType ? AxisValFromPixel(gName, xAxis , V_mouseX) : AxisValFromPixel(gName, xAxis , V_mouseY)
	
	Make/Free/T/N=(trNum) w_traces = StringFromList(p,traces)
	Make/Free/N=(trNum) w_order
	
	int i
	for(i = 0 ; i < trNum; i++)
		trace = StringFromList(i,traces)
		variable xOff = 0, yOff = 0, xScale = 0, yScale = 0, ywVal, ywMin
		getAllTraceOffsets(gName, trace, xOff, yOff, xScale, yScale)
		int err = getYfromWave(gName, trace, (xClick-xOff)/xScale, ywVal, ywMin)				// extract wave min and Y value at X position
		w_order[i] = err ? yOff : yOff+ywVal*yScale
	endfor
	
	if (GetKeyState(0) & 2^0)																	// holding ctrl / command triggers descending order sorting
		Sort/R w_order, w_order, w_traces
	else
		Sort   w_order, w_order, w_traces
	endif
	
	if (trNum <= 100)																			// this even works for instances (but is limited to 100 traces)
		wfprintf traces, "%s,", w_traces
		Execute/Q "ReorderTraces _front_,{"+RemoveEnding(traces,",")+"}"
	else
		for(i = 0; i < numpnts(w_traces); i++)
			ReorderTraces/W=$gName _front_, {$w_traces[i]}
		endfor
	endif
	return 0
end

//################################################################################################

static function OrderTracesAlphabetical()														// reorder traces based on trace names
	GetLastUserMenuInfo																			// figure out graph and trace names
	string gName = S_graphName, trace = S_traceName, traces = ""
	int trNum = FetchAllTraces(gName, traces)
	if (!trNum)
		return -1
	endif
	variable sortOpt = (GetKeyState(0) & 2^0) + 16												// holding ctrl / command triggers descending order sorting
	
	int i, do2Dsort = 0
	string instanceList = GrepList(traces,"#")
	if (ItemsInList(instanceList))
		if (WaveDims(TraceNameToWaveRef(gName, StringFromList(0, instanceList))) == 2)			// special treatment of 2D waterfall plots
			if (trNum <= 100)																	// this method is limited to 100 traces
				do2Dsort = 1
			else
				Print "Error: Cannot sort 2D data larger that 100 traces."
				return 0
			endif
		endif
	endif
	
	if (!do2Dsort)
		traces = SortList(traces, ";", sortOpt)
		for(i = 0; i < ItemsInList(traces); i++)
			ReorderTraces/W=$gName _front_, {$StringFromList(i,traces)}
		endfor
		return 0
	endif
	
	string curtrace, basename, instance, copylist = ""
	for(i = 0; i < ItemsInList(traces); i++)													// add range for sorting to list
		curtrace = StringFromList(i,traces)
		basename = StringFromList(0,curtrace,"#")
		instance = StringFromList(1,curtrace,"#")
		copylist += basename + "@@"+StringByKey("YRANGE",TraceInfo(gName,curtrace,0))+"@@" + instance + ";"
	endfor
	
	copylist = SortList(copylist, ";", sortOpt); traces = ""
	
	for(i = 0; i < ItemsInList(copylist); i++)													// remove range
		SplitString/E=("(.*)@@(.*)@@(.*)") StringFromList(i,copylist), basename, curtrace, instance
		traces += basename + SelectString(strlen(instance),"","#") + instance + ";"
	endfor
	Execute/Q "ReorderTraces _front_,{"+RemoveEnding(ReplaceString(";",traces,","),",")+"}"
	return 0
end

//################################################################################################

static function SpreadToFillRange()																// spread (equally offset) spectra over Y axis range; traces are stacked in the order of plotting
	string traces = "", gName = getTopGraph()
	variable trNum = FetchAllTraces(gName,traces)
	if (!trNum)
		return -1
	endif
	
	Make/Free/N=(trNum)/T w_traces = StringFromList(p,traces)
	Make/Free/N=(trNum) w_offsets = getTraceOffset(gName, w_traces[p], 1)
	variable yMin, yMax, low, high
	getMinMaxfromWave(gName, w_traces[0], yMin, yMax);		 low  = yMin + w_offsets[0]			// find low from first trace
	getMinMaxfromWave(gName, w_traces[trNum-1], yMin, yMax); high = yMax + w_offsets[trNum-1]	// find high from last trace
	
	GetAxis/W=$gName/Q $GetMajorAxisFromTraceList(gName, traces, "Y")
	variable bottomGap = low-v_Min, topGap = v_Max-high
	variable i, dOffset = (topGap+bottomGap)/(trNum-1)
	for (i = 0; i < trNum; i++)																	// apply offsets
		ModifyGraph/W=$gName offset($w_traces[i])={,w_offsets[i]- bottomGap + dOffset*i}
	endfor
	return 0
end

//################################################################################################

static function SpreadSpectra(int which)														// which: 0 = equal offset increment, 1 = equal spacing
	string traces = "", gName = getTopGraph()
	int trNum = FetchAllTraces(gName,traces)
	if (!trNum)
		return -1
	endif
	
	Make/Free/N=(trNum)/T w_traces = StringFromList(p,traces)
	Make/Free/N=(trNum)/wave w_waves = TraceNameToWaveRef(gName,w_traces)
	Make/Free/N=(trNum) w_low, w_high, w_range	
	variable i, yMin, yMax
	for (i = 0; i < trNum; i++)
		getMinMaxfromWave(gName, w_traces[i], yMin, yMax);  w_low[i] = yMin;  w_high[i] = yMax;
	endfor
	w_range = w_high-w_low
	
	GetAxis/W=$gName/Q $GetMajorAxisFromTraceList(gName, traces, "Y")
	variable yAxMin = v_min, yAxRange = (v_max-v_min), dRange = sum(w_range)
	variable gap = which ? (yAxRange-dRange)/(trNum-1) : (yAxRange-w_range[trNum-1])/(trNum-1)	// gap or offset between each trace (defined as distance between highest value of spectrum n and lowest value of spectrum n+1)
	variable newOff
	for (i = 0; i < trNum; i++)
		if (i == 0)
			newOff = yAxMin - w_low[0]															// add this to bring first trace down to bottom
		else
			newOff = which ? w_high[i-1] - w_low[i] + gap : w_low[i-1] - w_low[i] + gap
		endif
		ModifyGraph/W=$gName offset($w_traces[i])={,newOff}
		w_high[i] += newOff
		w_low[i]  += newOff
	endfor
	return 0
end

//################################################################################################

static function/S getTopMainGraph()
	return StringFromList(0,WinList("*", ";", "WIN:1"))
end

static function/S getTopGraph()
	string windows = WinList("*", ";", "")
	string topwin = StringFromList(0, windows)
	int i = 1
	do																							// try finding a valid graph, even if it is not the top one
		int wtype = WinType(topwin)
		if (wtype == 1 || wtype == 3 || wtype == 7)												// top window is graph or a type that can contain a graph
			GetWindow $topwin, activeSW															// for the active sub window (this can also be the top-level, if there are not children, or no child is active)
			if (strlen(S_value) > 0 && WinType(S_value) == 1)
				return S_value																	// this isn't necessarily an immediate child of top window
			elseif (wtype == 1)
				return topwin
			endif
		endif
		topwin = StringFromList(i, windows)
		i++
	while (strlen(topwin) > 0)
	return ""																					// no graph found
end

//________________________________________________________________________________________________
//
//										TRACE OFFSET PANEL
//________________________________________________________________________________________________

function BuildTraceOffsetPanel()
	if (WinType("TraceOffsetPanel"))
		DoWindow/F TraceOffsetPanel
		return 0
	endif
	
	variable pW = 224, pH = 450
	
	variable TFTResX, TFTResY																	// get the display resolution to align window better (center of full screen)
	if (cmpstr(IgorInfo(2),"Macintosh")==0)		// Mac
		string TFTInfo = StringByKey("RECT",ReplaceString(",RECT",StringByKey("SCREEN1",IgorInfo(0),":"),";RECT"),"=")
		TFTResX = (Str2Num(StringFromList(2,TFTInfo,",")) - Str2Num(StringFromList(0,TFTInfo,",")))*(72/ScreenResolution)  
		TFTResY = (Str2Num(StringFromList(3,TFTInfo,",")) - Str2Num(StringFromList(1,TFTInfo,",")))*(72/ScreenResolution)  
	else										// Windows
		GetWindow kwFrameInner wsizeDC
		TFTResX = (v_right-v_left)
		TFTResY = (v_bottom-v_top)
	endif
	
	NewPanel/K=1/W=((TFTResX-pW)/2,(TFTResY-pH)/2,(TFTResX+pW)/2,(TFTResY+pH)/2)/N=TraceOffsetPanel as "Modify Traces"
	ModifyPanel/W=TraceOffsetPanel ,fixedSize=1
	SetWindow TraceOffsetPanel hook(UpdateHook)=OffsetPanelUpdateFunc
	SetWindow TraceOffsetPanel userdata(procVersion)=num2str(kVersion)							// save procedure version

	TabControl offsetTabs ,pos={5,5} ,size={216,408} ,tabLabel(0)="All Traces" ,tabLabel(1)="Individual Trace" ,proc=TraceTool_TabControl
	TabControl offsetTabs focusring=0
	
	DefineGuide TabL={FL,5}
	DefineGuide TabR={FR,-5}
	DefineGuide TabT={FT,30}
	DefineGuide TabB={FB,-5}
	// +++++++++++++++++++++++++++++++++++ multi traces tab ++++++++++++++++++++++++++++++++++++++
	string pMulti = "TraceOffsetPanel#MultiTab"
	NewPanel/FG=(TabL, TabT, TabR, TabB)/HOST=#/N=MultiTab
	ModifyPanel frameStyle=0, frameInset=0
	ModifyPanel cbRGB=(60000,60000,60000,0)
	
	SetDrawLayer UserBack
	DrawLine 12,280,205,280
	DrawLine 12,345,205,345
	
	TitleBox MultiTargetTitle	,pos={10,3}		,size={195,18}	,title=""	,fixedSize=1		,frame=0
	
	CheckBox cHiddenTraces		,pos={13,25}	,size={60,22}	,title="Include Hidden Traces"	,value = 1		,help={"Include hidden traces when scaling."}
	SetVariable vaSetYOffset	,pos={13,70}	,size={90,16}	,title="y:"
	SetVariable vaScaleYOffset	,pos={13,110}	,size={90,16}	,title="y:"
	SetVariable vaShiftYOffset	,pos={13,150}	,size={90,16}	,title="y:"
	SetVariable vaSetXOffset	,pos={113,70}	,size={90,16}	,title="x:"
	SetVariable vaScaleXOffset	,pos={113,110}	,size={90,16}	,title="x:"
	SetVariable vaShiftXOffset	,pos={113,150}	,size={90,16}	,title="x:"
	
	CheckBox ChooseSet			,pos={13,50}	,size={60,22}	,title="Set or "				,value=1		,help={"Sets and overrides offsets of traces."}
	CheckBox ChooseAdd			,pos={67,50}	,size={60,22}	,title="Add"					,value=0		,help={"Adds to the current trace offsets."}
	TitleBox SetTitle			,pos={110,50}	,size={160,13}	,title="Offset per Trace"
	TitleBox ScaleTitle			,pos={13,93}	,size={160,13}	,title="Scale Offset per Trace"
	TitleBox ShiftTitle			,pos={13,133}	,size={160,13}	,title="Shift Offset of All Traces"
	Button ResetYButton			,pos={13,175}	,size={90,22}	,title="Y Reset"								,help={"Resets all Y offsets to 0."}
	Button ResetXButton			,pos={113,175}	,size={90,22}	,title="X Reset"								,help={"Resets all X offsets to 0."}
	Button ReverseButton		,pos={13,205}	,size={190,22}	,title="Reverse Sorting"						,help={"Reverses the order in which the traces are sorted in the graph."}
	SetVariable vTraceFilter	,pos={13,235}	,size={190,16}	,title="Filter:"								,help={"Apply offsets only to traces with matching name. You can use wildcard characters."}
	CheckBox cOmitFilter		,pos={47,257}	,size={90,22}	,title="Filter Excludes Traces"	,value = 0		,help={"The filter keyword is used to omit traces instead of including them."}

	PopupMenu ColorSelect		,pos={13,290}	,size={190,22}									,bodyWidth=190
	Button ColorButton			,pos={13,315}	,size={110,22}	,title="Colorize"								,help={"Colorize traces with currently selected color spectrum. To reverse spectrum hold 'ctrl'/'command'."}
	CheckBox cReverseColor		,pos={133,318}	,size={60,22}	,title="in Reverse"				,value = 0		,help={"Applies the color in reverse (also works directly by holding 'ctrl'/'command' while pressing the button.)"}
	PopUpMenu ProfileSelect		,pos={13,353}	,size={190,20}	,title="Profile"				,bodywidth=152
	
	Button vaApplyAllButton		,pos={5,387}	,size={205,23}	,title="Apply All to Data (overwrite!)"			,help={"Applies all offsets and scales to 1D data. Will overwrite the original waves, so use this with care!"}
	// ++++++++++++++++++++++++++++++++++++++++ settings +++++++++++++++++++++++++++++++++++++++++
	SetVariable vaSetYOffset	,value=_NUM:0	,limits={0,inf,1}					,userdata="0"
	SetVariable vaScaleYOffset	,value=_NUM:1	,limits={0,inf,kMultiIncrStep}		,userdata="1"
	SetVariable vaShiftYOffset	,value=_NUM:0	,limits={-inf,inf,1}
	SetVariable vaSetXOffset	,value=_NUM:0	,limits={-inf,inf,0.1}				,userdata="0"
	SetVariable vaScaleXOffset	,value=_NUM:1	,limits={0,inf,kMultiIncrStep}		,userdata="1"
	SetVariable vaShiftXOffset	,value=_NUM:0	,limits={-inf,inf,1}
	SetVariable vTraceFilter	,value=_STR:"*"								,proc=TraceTool_TraceFilterProc
	
	PopupMenu ColorSelect		,mode=2	,value="*COLORTABLEPOPNONAMES*"							// must be mode first
	PopupMenu ProfileSelect		,mode=1	,value=TraceTool_ProfileWaveList()	,proc=TraceTool_ProfileSelect
	
	ModifyControlList ControlNameList(pMulti,";","Choose*")	,win=$pMulti	,proc=TraceTool_SwitchOffsetMode,mode = 1
	ModifyControlList ControlNameList(pMulti,";","*Button")	,win=$pMulti	,proc=TraceTool_ButtonFunctions
	ModifyControlList ControlNameList(pMulti,";","vaS*") 	,win=$pMulti	,proc=TraceTool_MultiTraceVars	,format="%g"
	ModifyControlList ControlNameList(pMulti,";","*Title")	,win=$pMulti	,frame=0
	ModifyControlList ControlNameList(pMulti,";","vaS*")+"ColorSelect;ProfileSelect;" ,win=$pMulti ,focusRing=0
	
	if (CmpStr(IgorInfo(2), "Macintosh") == 0)
		ModifyControlList ControlNameList(pMulti,";","*")	,win=$pMulti	,fsize=10
	endif
	SetActiveSubwindow ##
	
	// ++++++++++++++++++++++++++++++++++++ single trace tab +++++++++++++++++++++++++++++++++++++
	string pSingle = "TraceOffsetPanel#SingleTab"
	NewPanel/FG=(TabL, TabT, TabR, TabB)/HOST=#/N=SingleTab
	ModifyPanel frameStyle=0, frameInset=0
	ModifyPanel cbRGB=(60000,60000,60000,0)
	
	SetDrawLayer UserBack
	DrawLine 12,235,205,235
	DrawLine 12,317,205,317

	TitleBox SingleTargetTitle	,pos={10,3}		,size={195,18}	,title=""			,frame=0	,fixedSize=1

	Button CsrActButton			,pos={10,25}	,size={60,22}	,title="Cursor:"	,fstyle=1					,help={"Sets a new cursor on the graph."}
	CheckBox traceCsrA			,pos={78,28}	,size={60,22}	,title="A"										,help={"Selects cursor A (a key)."}
	CheckBox traceCsrB			,pos={111,28}	,size={60,22}	,title="B"										,help={"Selects cursor B (b key)."}
	CheckBox traceCsrC			,pos={144,28}	,size={60,22}	,title="C"										,help={"Selects cursor C (c key)."}
	CheckBox traceCsrD			,pos={177,28}	,size={60,22}	,title="D"										,help={"Selects cursor D (d key)."}
	
	TitleBox trSetTitle			,pos={13,53}	,size={170,13}	,title="Set Trace Offset"
	TitleBox trScaleTitle		,pos={13,93}	,size={170,13}	,title="Set Trace Multiplier"
	TitleBox trShiftTitle		,pos={13,133}	,size={170,13}	,title="Reposition at Cursor"
	SetVariable trSetYOffset	,pos={13,70}	,size={90,16}	,title="y:"
	SetVariable trScaleYOffset	,pos={13,110}	,size={90,16}	,title="y:"
	SetVariable trShiftYOffset	,pos={13,150}	,size={90,16}	,title="y:"
	SetVariable trSetXOffset	,pos={113,70}	,size={90,16}	,title="x:"
	SetVariable trScaleXOffset	,pos={113,110}	,size={90,16}	,title="x:"
	SetVariable trShiftXOffset	,pos={113,150}	,size={90,16}	,title="x:"
	SetVariable trScaleNorm		,pos={13,178}	,size={140,16}	,title="Normalize:"				,bodyWidth=80	,help={"Adjusts the multiplier to normalize the trace height to this value at the current cursor's position (excluding the Y offset)."}
	Button trNormTo1Button		,pos={163,177}	,size={40,20}	,title="to 1"									,help={"Normalizes trace height to 1 at current cursor position (excluding the Y offset)."}
	Button trResetYButton		,pos={13,205}	,size={90,22}	,title="Y Reset"								,help={"Resets Y offset and multiplier to 0."}
	Button trResetXButton		,pos={113,205}	,size={90,22}	,title="X Reset"								,help={"Resets X offset and multiplier to 0."}
	
	TitleBox trCsrControlTitle	,pos={13,240}	,size={190,13}	,title="Control Cursor Position"
	Button trPrevTraceButton	,pos={13,258}	,size={90,22}	,title="< Prev Trace"							,help={"Moves the cursor to onto the previous trace in the list of traces (page down key)."}
	Button trNextTraceButton	,pos={113,258}	,size={90,22}	,title="Next Trace >"							,help={"Moves the cursor to onto the next trace in the list of traces (page up key)."}
	Button trGoToMaxButton		,pos={13,288}	,size={90,22}	,title="Go to Max"								,help={"Sets the cursor to trace's maximum (h key)."}
	Button trGoToMinButton		,pos={113,288}	,size={90,22}	,title="Go to Min"								,help={"Sets the cursor to trace's minimum (l key)."}
	
	SetVariable trOffxDelta		,pos={58,325}	,size={100,16}	,title="dx:"					,bodyWidth=80	,help={"Negative values will flip trace in X direction."}
	SetVariable trOffyDelta		,pos={58,353}	,size={100,16}	,title="dy:"					,bodyWidth=80	,help={"Negative values will flip trace in Y direction."}
	Button trAddXButton			,pos={168,323}	,size={35,20}	,title="+"			,fstyle=1	,fsize=14		,help={"Add X delta value."}
	Button trRemXButton			,pos={13,323}	,size={35,20}	,title="-"			,fstyle=1	,fsize=14		,help={"Subtract X delta value."}
	Button trAddYButton			,pos={168,351}	,size={35,20}	,title="+"			,fstyle=1	,fsize=14		,help={"Add Y delta value."}
	Button trRemYButton			,pos={13,351}	,size={35,20}	,title="-"			,fstyle=1	,fsize=14		,help={"Subtract Y delta value."}

	Button trApplyCurButton		,pos={5,387}	,size={205,23}	,title="Apply to Current Data (overwrite!)"		,help={"Applies offsets and scales to the currently selected 1D data. Will overwrite the original waves, so use this with care!"}
	// ++++++++++++++++++++++++++++++++++++++++ settings +++++++++++++++++++++++++++++++++++++++++
	SetVariable trSetYOffset	,value=_NUM:0	,limits={-inf,inf,1}
	SetVariable trScaleYOffset	,value=_NUM:1	,limits={-inf,inf,kMultiIncrStep}
	SetVariable trShiftYOffset	,value=_NUM:0	,limits={-inf,inf,0}
	SetVariable trSetXOffset	,value=_NUM:0	,limits={-inf,inf,0.1}
	SetVariable trScaleXOffset	,value=_NUM:1	,limits={-inf,inf,kMultiIncrStep}
	SetVariable trShiftXOffset	,value=_NUM:0	,limits={-inf,inf,0}
	SetVariable trScaleNorm		,value=_NUM:0	,limits={-inf,inf,1}
	
	SetVariable trOffxDelta		,value=_NUM:0	,limits={-inf,inf,0}
	SetVariable trOffyDelta		,value=_NUM:0	,limits={-inf,inf,0}
	
	ModifyControlList ControlNameList(pSingle,";","tr*")		,win=$pSingle	,disable=2		// disable all cursor related controls for now
	ModifyControlList ControlNameList(pSingle,";","traceCsr*")	,win=$pSingle	,proc=TraceTool_CursorSelector	,mode = 1
	ModifyControlList ControlNameList(pSingle,";","*Button")	,win=$pSingle	,proc=TraceTool_ButtonFunctions
	ModifyControlList ControlNameList(pSingle,";","trS*") 		,win=$pSingle	,proc=TraceTool_SingleTraceVars	,format="%g" ,focusRing=0
	ModifyControlList ControlNameList(pSingle,";","*Title")		,win=$pSingle	,frame=0
	
	if (CmpStr(IgorInfo(2), "Macintosh") == 0)
		ModifyControlList ControlNameList(pSingle,";","*")		,win=$pSingle	,fsize=10
	endif
	SetActiveSubwindow ##
	
	int csrSet = 0;	string gName = getTopGraph()
	if (strlen(gName))																			// check whether one of the cursors is set
		csrSet = strlen(CsrInfo(A,gName)+CsrInfo(B,gName)+CsrInfo(C,gName)+CsrInfo(D,gName)) > 0
		AutoPositionWindow/M=0/R=$gName TraceOffsetPanel
	endif
	TabControl offsetTabs value=csrSet															// if a cursor is set, display the single trace tab
	TraceTool_ToggleTabs(csrSet)
	FetchLatestTraceValues()																	// set variable values from top window
	return 0
end
//+++++++++++++++++++++++++++++++++++++ popup help functions +++++++++++++++++++++++++++++++++++++
function/S TraceTool_ProfileWaveList()
	return "none;"+WaveList("*",";", "DIMS:1")
end

//################################################################################################

static function getAllTraceOffsets(string gName, string trace, variable &xOff, variable &yOff, variable &xScale, variable &yScale)
	sscanf StringByKey("offset(x)",   TraceInfo(gName,trace,0), "="), "{%f,%f}", xOff, yOff
	sscanf StringByKey("muloffset(x)",TraceInfo(gName,trace,0), "="), "{%f,%f}", xScale, yScale
	yScale = yScale == 0? 1 : yScale
	xScale = xScale == 0? 1 : xScale
end

//################################################################################################

static function getTraceOffset(string gName, string trace, int axis)
	variable xOff = 0, yOff = 0
	sscanf StringByKey("offset(x)", TraceInfo(gName,trace,0), "="), "{%f,%f}", xOff, yOff
	return axis ? yOff : xOff
end

//################################ panel update helper function ##################################

function OffsetPanelUpdateFunc(STRUCT WMWinHookStruct &s)
	int HookTakeover = 0
	StrSwitch (s.EventName)
		case "activate":
			HookTakeover = 1
			if (!CmpStr(s.winName,"OffsetPanel"))												// in case an older panel is open -> rebuild
				KillWindow/Z OffsetPanel
				BuildTraceOffsetPanel()
			endif
			FetchLatestTraceValues()
		break
		case "keyboard":
			string gName = getTopGraph()
			if (!strlen(gName))
				return 0
			endif
			HookTakeover = 1
			string key = ""
			Switch(s.keycode)
				case 11:	// page up
					key = "trNextTraceButton"
				break
				case 12:	// page down
					key = "trPrevTraceButton"
				break
				case 104:	// h
					key = "trGoToMaxButton"
				break
				case 108:	// l
					key = "trGoToMinButton"
				break
			EndSwitch
			TraceTool_ExecuteButtonAction(key)
			
			if (s.keycode > 96 && s.keycode < 101)	// a-d keys
				key = StringFromList(s.keycode-97, "A;B;C;D;")
				if (strlen(CsrInfo($key,gName)))
					TraceTool_SwitchCurrentCursor("traceCsr"+key)
				endif
			endif
			
			if (s.keycode > 27 && s.keycode < 32)	// left, right, up, down
				HookTakeover = TraceTool_KeypadShiftandScale(s.keycode)
			endif
		break
	EndSwitch
	return HookTakeover
end

//################################################################################################

function TraceTool_ToggleTabs(int whichTab)
	SetWindow TraceOffsetPanel#MultiTab  hide=whichTab==1
	SetWindow TraceOffsetPanel#SingleTab hide=whichTab==0
end

function TraceTool_TabControl(STRUCT WMTabControlAction &s) : TabControl
	if (s.eventCode == 2)
		TraceTool_ToggleTabs(s.tab)
	endif
end

//################################################################################################

function TraceTool_KeypadShiftandScale(int key)
	string gName = getTopGraph()
	if (!strlen(gName))
		return 0
	endif
	string trace = GetUserData("TraceOffsetPanel", "", "trace")									// get current trace name and cursor from panel user data
	string csr   = GetUserData("TraceOffsetPanel", "", "cursor")
	if (strlen(trace) == 0 || strlen(csr) == 0)
		return 0
	endif
	
	variable xOff = 0, yOff = 0, xScale = 0, yScale = 0
	getAllTraceOffsets(gName, trace, xOff, yOff, xScale, yScale)
	variable yScaleMult = kMultiIncrStep*10^GetValuesOrderOfMagnitude(yScale) 
	variable xScaleMult = kMultiIncrStep*10^GetValuesOrderOfMagnitude(xScale) 
	
	wave w = TraceNameToWaveRef(gName, trace)
	wave/Z w_x = XWaveRefFromTrace(gName, trace)
	
	WaveStats/Q w
	variable yDelta = V_max*yScale*10*kOffsetIncrStep
	variable xDelta = DimDelta(w,0), mulDelta													// delta shift values for cursor control
	if (WaveExists(w_x))
		xDelta = abs(leftx(w_x)-rightx(w_x))/(numpnts(w_x))
	endif
	
	variable neg = 1, ctrlKey = (GetKeyState(0) & 2^0)!=0
	Switch(key)
		case 28:	// left
			neg = -1
		case 29:	// right
			if (ctrlKey)
				mulDelta = xScale*neg == -xScaleMult/5 ? xScaleMult*2/5 : xScaleMult/5			// don't scale to exactly zero (which will flip back to 1)
				ModifyGraph/W=$gName muloffset($trace)={xScale+neg*mulDelta,}
			else
				ModifyGraph/W=$gName offset($trace)={xOff+neg*xDelta,}
			endif
		break
		case 31:	// down
			neg = -1
		case 30:	// up
			if (ctrlKey)
				mulDelta = yScale*neg == -yScaleMult/5 ? yScaleMult*2/5 : yScaleMult/5
				ModifyGraph/W=$gName muloffset($trace)={,yScale+neg*mulDelta}
			else
				ModifyGraph/W=$gName offset($trace)={,yOff+neg*yDelta}
			endif
		break
	EndSwitch
	FetchLatestCursorValues(csr)
	AdjustProfileAxisScaling()
	return 1
end

//################################################################################################

function TraceTool_SwitchOffsetMode(STRUCT WMCheckboxAction &s) : CheckBoxControl
	if(s.eventCode == 2)
		CheckBox $StringFromList(0,RemoveFromList(s.ctrlName, ControlNameList("TraceOffsetPanel#MultiTab",";","Choose*"))) ,win=TraceOffsetPanel#MultiTab ,value=0		
	endif
	return 0
end

//################################################################################################

function TraceTool_CursorSelector(STRUCT WMCheckboxAction &s) : CheckBoxControl
	if(s.eventCode == 2)
		TraceTool_SwitchCurrentCursor(s.ctrlName)
	endif
	return 0
end

//################################################################################################

function TraceTool_SwitchCurrentCursor(string which)
	int i;	string ctrl = RemoveFromList(which, ControlNameList("TraceOffsetPanel#SingleTab",";","traceCsr*"))
	for (i = 0; i < ItemsInList(ctrl); i++)
		CheckBox $StringFromList(i,ctrl) ,win=TraceOffsetPanel#SingleTab ,value=0
	endfor
	CheckBox $which ,win=TraceOffsetPanel#SingleTab ,value=1
	FetchLatestCursorValues(ReplaceString("traceCsr", which, ""))
	return 0
end

//################################################################################################

static function FetchLatestTraceValues()														// update both left and right side values (if cursors are set)
	string gName = getTopGraph()
	string pSingle = "TraceOffsetPanel#SingleTab"
	string pMulti  = "TraceOffsetPanel#MultiTab"
	
	ModifyControlList/Z ControlNameList(pSingle,";","tr*")+"trApplyCurButton;",win=$pSingle ,disable=2	// disable cursor related controls for now
	if (!strlen(gName))
		TitleBox MultiTargetTitle  ,win=$pMulti  ,title="No Graph Available!" ,fColor=(65535,0,0)
		TitleBox SingleTargetTitle ,win=$pSingle ,title="No Graph Available!" ,fColor=(65535,0,0)
		return 0
	endif
	string traces = ""
	variable items = FetchTraces(gName,traces)
	if (items == 0)
		return -1
	endif
	
	variable xOff = 0, yOff = 0, xShift = 0,  yShift = 0
	sscanf StringByKey("offset(x)", TraceInfo(gName,StringFromList(items-1,traces),0), "="), "{%f,%f}", xShift, yShift
	if (items > 1)																				// calculate shift and offset from first two traces
		sscanf StringByKey("offset(x)", TraceInfo(gName,StringFromList(items-2,traces),0), "="), "{%f,%f}", xOff, yOff
	else
		xOff = xShift; yOff = yShift
	endif
	
	variable xRange, yRange;	string xUnit = "", yUnit = ""
	GetAxisRangeFromTraceList(xRange, yRange, xUnit, yUnit, gName, traces)
	TitleBox MultiTargetTitle 	,win=$pMulti	,title="\f01Modify:\f00 "+gName ,fColor=(0,0,0)
	SetVariable vaSetYOffset	,win=$pMulti	,value=_NUM:yOff-yShift	,format="%g"+yUnit	,limits={-inf,inf,yRange*kOffsetIncrStep}	,userdata=num2str(yOff-yShift)
	SetVariable vaShiftYOffset	,win=$pMulti	,value=_NUM:yShift		,format="%g"+yUnit	,limits={-inf,inf,yRange*kOffsetIncrStep}
	SetVariable vaSetXOffset	,win=$pMulti	,value=_NUM:xOff-xShift	,format="%g"+xUnit	,limits={-inf,inf,xRange*kOffsetIncrStep}	,userdata=num2str(xOff-xShift)
	SetVariable vaShiftXOffset	,win=$pMulti	,value=_NUM:xShift		,format="%g"+xUnit	,limits={-inf,inf,xRange*kOffsetIncrStep*10}
	PopupMenu ProfileSelect 	,win=$pMulti 	,mode=1											// update the selected profile or select none
	if (FindListItem("Profileleft", AxisList(gName)) >= 0)
		wave profile = TraceNameToWaveRef(gName, "AddTraceProfile")
		PopupMenu ProfileSelect	,win=$pMulti	,popmatch=nameofwave(profile)
	endif
	
	wave data = TraceNameToWaveRef(gName, StringFromList(0,traces))
	if (!WaveExists(data))																		// probably image data => not useful here
		return 0
	endif
	Button vaApplyAllButton 	,win=$pMulti 	,disable=2*(WaveDims(data) != 1)				// this button only works for 1D traces
	//------------------------------------- cursor controls --------------------------------------
	string csr = "A;B;C;D;", currCsr															// fetch available cursor list
	variable i, selCsr = -1,  setCsr = -1, csrActive = 0
	for (i = 0; i < ItemsInList(csr); i++)														// find currently selected cursor
		currCsr = "traceCsr"+StringFromList(i,csr)
		ControlInfo/W=$pSingle $currCsr
		if (V_Value)
			CheckBox $currCsr ,win=$pSingle ,value=0											// deselect for now
			selCsr = i
			break
		endif
	endfor

	for (i = 0; i < ItemsInList(csr); i++)
		currCsr = StringFromList(i,csr)
		if (strlen(CsrInfo($currCsr,gName)))
			if (numtype(zcsr($currCsr,gName)) != 2)												// check for cursors on image plots
				continue
			endif
			if (csrActive == 0 && selCsr == i)													// the currently selected cursor is active and can stay
				csrActive = 1
				setCsr = i
			endif
			if (setCsr == -1)																	// find the first free cursor from left
				setCsr = i
			endif
			CheckBox $("traceCsr"+currCsr) ,win=$pSingle ,disable=0
		endif
	endfor
	
	if (setCsr == -1)
		SetWindow TraceOffsetPanel	,userdata(cursor)=""										// clear cursor user data
		if (selCsr != -1)
			CheckBox $("traceCsr"+StringFromList(selCsr,csr)) ,win=$pSingle ,value=1			// set previous cursor (disabled but keeps cursor selected)
		endif
		TitleBox SingleTargetTitle	,win=$pSingle ,title="No Trace Selected!" ,fColor=(65535,0,0)
		return 0
	endif
	//---------------------------------- if cursor was found -------------------------------------
	currCsr = StringFromList(setCsr,csr)
	CheckBox $("traceCsr"+currCsr) ,win=$pSingle ,value=1
	string ctrlList = RemoveFromList("traceCsrA;traceCsrB;traceCsrC;traceCsrD;",ControlNameList(pSingle,";","tr*"))+"trApplyCurButton;"
	ModifyControlList ctrlList	,win=$pSingle	,disable=0
	FetchLatestCursorValues(currCsr)															// now get cursor specific values (right side)
	return 0
end

//################################################################################################

static function FetchLatestCursorValues(string csr [,int keepDelta])
	string gName = getTopGraph()
	string pSingle = "TraceOffsetPanel#SingleTab"
	if (!strlen(gName))
		return 0
	endif
	
	int updateOffsets = ParamIsDefault(keepDelta) ? 1 : !keepDelta
	
	string trace = StringByKey("TNAME", CsrInfo($csr,gName))									// get the info from the trace-associated axis
	string xAxis = StringByKey("XAXIS", TraceInfo(gName,trace,0))
	string yAxis = StringByKey("YAXIS", TraceInfo(gName,trace,0))
	string color = StringByKey("rgb(x)",TraceInfo(gName,trace,0), "=")							// show the color of the selected trace
	
	string xUnit = "", yUnit = ""
	GetAxis/W=$gName/Q $yAxis;	variable yRange = abs(V_max - V_min)
	yUnit = StringByKey("UNITS", AxisInfo(gName,yAxis))
	yUnit = SelectString(strlen(yUnit),""," "+yUnit)
	GetAxis/W=$gName/Q $xAxis;	variable xRange = abs(V_max - V_min)
	xUnit = StringByKey("UNITS", AxisInfo(gName,xAxis))
	xUnit = SelectString(strlen(xUnit),""," "+xUnit)
	
	variable xOff = 0, yOff = 0, xScale = 0, yScale = 0, shift
	getAllTraceOffsets(gName, trace, xOff, yOff, xScale, yScale)
	variable scaleNorm = vcsr($csr, gName)*yScale
	variable xShiftCsr = round( (hcsr($csr, gName)*xScale+xOff) *10^6)/10^6						// round to prevent floating point errors
	variable yShiftCsr = round( (vcsr($csr, gName)*yScale+yOff) *10^6)/10^6
	variable yScOrder  = GetValuesOrderOfMagnitude(yScale)
	variable xScOrder  = GetValuesOrderOfMagnitude(xScale) 
	
	TitleBox SingleTargetTitle	,win=$pSingle	,title="\f01Trace: \K"+color+ReplaceString("'",trace,"")+"\K(0,0,0)\f00",fColor=(0,0,0)
	Button trApplyCurButton 	,win=$pSingle	,disable=2*(WaveDims(TraceNameToWaveRef(gName,trace))!=1)		// this button only works for 1D traces
	SetVariable trSetYOffset	,win=$pSingle	,value=_NUM:yOff		,format="%g"+yUnit	,limits={-inf,inf,yRange*kOffsetIncrStep}
	SetVariable trShiftYOffset	,win=$pSingle	,value=_NUM:yShiftCsr	,format="%g"+yUnit	,limits={-inf,inf,yRange*kOffsetIncrStep*10}
	SetVariable trScaleYOffset	,win=$pSingle	,value=_NUM:yScale		,format="%g"		,limits={-inf,inf,kMultiIncrStep*10^yScOrder}
	SetVariable trSetXOffset	,win=$pSingle	,value=_NUM:xOff		,format="%g"+xUnit	,limits={-inf,inf,xRange*kOffsetIncrStep}
	SetVariable trShiftXOffset	,win=$pSingle	,value=_NUM:xShiftCsr	,format="%g"+xUnit	,limits={-inf,inf,xRange*kOffsetIncrStep*10}
	SetVariable trScaleXOffset	,win=$pSingle	,value=_NUM:xScale		,format="%g"		,limits={-inf,inf,kMultiIncrStep*10^xScOrder}
	SetVariable trScaleNorm		,win=$pSingle	,value=_NUM:scaleNorm	,format="%g"+yUnit
	
	SetVariable trOffxDelta		,win=$pSingle							,format="%g"			// first reset unit
	SetVariable trOffyDelta		,win=$pSingle							,format="%g"
	
	string unit = ""
	shift = TraceTool_FetchXShift(TraceNameToWaveRef(gName,trace),unit)
if (!numtype(shift) && updateOffsets)
	SetVariable trOffxDelta		,win=$pSingle	,value=_NUM:shift		,format="%g"+SelectString(strlen(unit),""," "+unit)
endif
	unit = ""
	shift = TraceTool_FetchYShift(TraceNameToWaveRef(gName,trace),unit)
if (!numtype(shift) && updateOffsets)
	SetVariable trOffyDelta		,win=$pSingle	,value=_NUM:shift		,format="%g"+SelectString(strlen(unit),""," "+unit)
endif
	
	SetWindow TraceOffsetPanel, userdata(trace)=trace
	SetWindow TraceOffsetPanel, userdata(cursor)=csr
	return 0
end

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static function/S GetMajorAxisFromTraceList(string gName, string traces, string which)
	int i,j,items = ItemsInList(traces)
	Make/Free/T/N=(items) Axes = StringByKey(UpperStr(which)+"AXIS", TraceInfo(gName,StringFromList(p,traces),0))
	string axis = Axes[0]
	if (DimSize(Axes,0) > 1)
		FindDuplicates/Free/RT=Axes_clean/DT=Axes_dup Axes
		if (DimSize(Axes_clean,0) > 1)															// need to find the major axis
			Make/Free/I/N=(DimSize(Axes_clean,0)) tempCount
			for (i = 0; i < DimSize(Axes_clean,0); i++)
				for (j = 0; j < DimSize(Axes_dup,0); j++)
					tempCount[i] += StringMatch(Axes_clean[i],Axes_dup[j])						// count the number of duplicates
				endfor
			endfor
			WaveStats/Q tempCount
			return Axes_clean[V_maxRowLoc]														// the maximum amount of occurrences will be the main axis
		else
			return Axes_clean[0]
		endif
	else
		return Axes[0]
	endif
end

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static function GetAxisRangeFromTraceList(variable &xRange, variable &yRange, string &xUnit, string &yUnit, string gName, string traces)
	string info
	string xAxis = GetMajorAxisFromTraceList(gName, traces, "X")
	string yAxis = GetMajorAxisFromTraceList(gName, traces, "Y")
	GetAxis/W=$gName/Q $xAxis
	xRange = abs(V_max - V_min)
	xUnit  = StringByKey("UNITS", AxisInfo(gName,xAxis))
	xUnit  = SelectString(strlen(xUnit),""," "+xUnit)
	GetAxis/W=$gName/Q $yAxis
	yRange = abs(V_max - V_min)
	yUnit  = StringByKey("UNITS", AxisInfo(gName,yAxis))
	yUnit  = SelectString(strlen(yUnit),""," "+yUnit)
	sprintf info, "%.*g\r", 1, yRange;	yRange = str2num(info)									// rounding
	sprintf info, "%.*g\r", 1, xRange;	xRange = str2num(info)
	return 0
end

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static function GetValuesOrderOfMagnitude(variable val)
	string numChar = "", eChar = ""; sprintf numChar, "%.1E", val
	SplitString/E=("[0-9]+.[0-9]([eE][-+]?[0-9]+)?") numChar, eChar
	variable OrderOfMagnitude = str2num(eChar[1,strlen(eChar)])
	return numtype(OrderOfMagnitude)!=0 ? 0 : OrderOfMagnitude
end

//################################## variable helper function ####################################

function TraceTool_SingleTraceVars(STRUCT WMSetVariableAction &s) : SetVariableControl
	if (s.eventCode != 1 && s.eventCode != 2)
		return 0
	endif
	
	s.blockReentry=1
	string gName = getTopGraph()
	if (!strlen(gName))
		return 0
	endif
	string trace = GetUserData("TraceOffsetPanel", "", "trace")									// get current trace name and cursor from panel user data
	string csr   = GetUserData("TraceOffsetPanel", "", "cursor")
	if (strlen(trace) == 0 || strlen(csr) == 0)
		return 0
	endif
	
	variable xOff = 0, yOff = 0, xScale = 0, yScale = 0
	getAllTraceOffsets(gName, trace, xOff, yOff, xScale, yScale)
	if (StringMatch(s.ctrlName,"trScale*"))
		s.dval = abs(s.dval) < 10e-8 ? sign(s.dval)*10e-8 : s.dval								// prevent input from getting very small
	endif
	
	StrSwitch(s.ctrlName)
		case "trSetYOffset":
			ModifyGraph/W=$gName offset($trace)={,s.dval}
		break
		case "trSetXOffset":
			ModifyGraph/W=$gName offset($trace)={s.dval,}
		break
		case "trShiftYOffset":
			ModifyGraph/W=$gName offset($trace)={,(s.dval-vcsr($csr, gName)*yScale)}
		break
		case "trShiftXOffset":
			ModifyGraph/W=$gName offset($trace)={(s.dval-hcsr($csr, gName)*xScale),}
		break
		case "trScaleYOffset":
			ModifyGraph/W=$gName muloffset($trace)={,s.dval}
		break
		case "trScaleXOffset":
			ModifyGraph/W=$gName muloffset($trace)={s.dval,}
		break
		case "trScaleNorm":
			if (vcsr($csr, gName) != 0)
				ModifyGraph/W=$gName muloffset($trace)={,s.dval/vcsr($csr, gName)}
			endif
		break
	EndSwitch	
	FetchLatestCursorValues(csr)
	AdjustProfileAxisScaling()
	return 0
end

//################################################################################################

function TraceTool_MultiTraceVars(STRUCT WMSetVariableAction &s) : SetVariableControl
	if (s.eventCode != 1 && s.eventCode != 2)
		return 0
	endif
	
	s.blockReentry=1
	string gName = getTopGraph()
	if (!strlen(gName))
		return 0
	endif
	string traces = ""
	variable items = FetchTraces(gName,traces)
	if (items == 0)
		return 0
	endif
	
	variable prevVal, curXOff = 0, curYOff = 0, xOff = 0, yOff = 0, YnotX = 0, i
	sscanf StringByKey("offset(x)", TraceInfo(gName,StringFromList(items-1,traces),0), "="), "{%f,%f}", xOff, yOff
	StrSwitch(s.ctrlName)
		case "vaSetYOffset":
			YnotX = 1
		case "vaSetXOffset":
			ControlInfo/W=TraceOffsetPanel#MultiTab ChooseAdd;	int addMode = v_Value
			prevVal = str2num(s.Userdata);	s.Userdata = num2str(s.dval)						// get previous offset from and write current value to control user data
			for (i = 0; i < items; i++)
				if (addMode)
					sscanf StringByKey("offset(x)", TraceInfo(gName,StringFromList(items-1 - i,traces),0), "="), "{%f,%f}", curXOff, curYOff
					if (YnotX)
						ModifyGraph/W=$gName offset($StringFromList(items-1 - i,traces))={,i*(s.dval-prevVal)+curYOff}
					else
						ModifyGraph/W=$gName offset($StringFromList(items-1 - i,traces))={i*(s.dval-prevVal)+curXOff,}
					endif
				else
					if (YnotX)
						ModifyGraph/W=$gName offset($StringFromList(items-1 - i,traces))={,i*s.dval+yOff}
					else
						ModifyGraph/W=$gName offset($StringFromList(items-1 - i,traces))={i*s.dval+xOff,}
					endif
				endif
			endfor
		break
		case "vaScaleYOffset":
			YnotX = 1
		case "vaScaleXOffset":
			prevVal = str2num(s.Userdata);	s.Userdata = num2str(s.dval)						// get previous multiplier from and write current value to control user data
			prevVal = prevVal != 0 ? prevVal : 1												// make sure multiplier is not zero
			for (i = 0; i < items; i++)
				sscanf StringByKey("offset(x)", TraceInfo(gName,StringFromList(items-1 - i,traces),0), "="), "{%f,%f}", curXOff, curYOff
				if (YnotX)
					ModifyGraph/W=$gName offset($StringFromList(items-1 - i,traces))={,(curYOff-yOff)*s.dval/prevVal+yOff}
				else
					ModifyGraph/W=$gName offset($StringFromList(items-1 - i,traces))={(curXOff-xOff)*s.dval/prevVal+xOff,}
				endif
			endfor
		break
		case "vaShiftYOffset":
			YnotX = 1
		case "vaShiftXOffset":
			for (i = 0; i < items; i++)
				sscanf StringByKey("offset(x)", TraceInfo(gName,StringFromList(items-1 - i,traces),0), "="), "{%f,%f}", curXOff, curYOff
				if (YnotX)
					ModifyGraph/W=$gName offset($StringFromList(items-1 - i,traces))={,(curYOff-yOff)+s.dval}
				else
					ModifyGraph/W=$gName offset($StringFromList(items-1 - i,traces))={(curXOff-xOff)+s.dval,}
				endif
			endfor
		break
	EndSwitch
	string csr = GetUserData("TraceOffsetPanel", "", "cursor")
	if (strlen(csr))
		FetchLatestCursorValues(csr)
	endif
	AdjustProfileAxisScaling()
	return 0
end

// ###############################################################################################

function TraceTool_TraceFilterProc(STRUCT WMSetVariableAction &s) : SetVariableControl
	if (s.eventMod & 16 && s.eventCode == 9)													// add right-click menu to directly insert values
		string gName = getTopGraph()
		string traces = "(No traces found;", currStr
		
		ControlInfo/W=TraceOffsetPanel#MultiTab cHiddenTraces;	int noHidden = V_flag ? !V_Value : 0
		if (strlen(gName))
			traces = "(Add trace:;" + TraceNameList(gName,";",1+4*noHidden)
		endif
		
		PopupContextualMenu traces
		if (V_Flag > 1)
			ControlInfo/W=TraceOffsetPanel#MultiTab vTraceFilter
			currStr = SelectString(strlen(S_Value),"",RemoveEnding(S_Value,";")+";")
			currStr = SelectString(CmpStr(currStr,"*;"),"",currStr)
			SetVariable vTraceFilter ,win=TraceOffsetPanel#MultiTab ,value=_STR:currStr+S_Selection+";"
			ControlUpdate/W=TraceOffsetPanel#MultiTab vTraceFilter
		endif
		return 1
	endif
	return 0
end

//++++++++++++++++++++++++++++++ help function to get a trace list ++++++++++++++++++++++++++++++

static function FetchTraces(string win, string &traces)
	int noHidden = 0, omit = 0
	string filter = "", tempList = ""
	if (WinType("TraceOffsetPanel"))															// read settings from panel if open
		ControlInfo/W=TraceOffsetPanel#MultiTab cHiddenTraces
		if (V_flag)
			noHidden = !V_Value
		endif
		ControlInfo/W=TraceOffsetPanel#MultiTab vTraceFilter
		if (V_flag)
			filter = S_Value
		endif
		ControlInfo/W=TraceOffsetPanel#MultiTab cOmitFilter
		if (V_flag)
			omit = V_Value
		endif
	endif
	
	traces = TraceNameList(win,";",1+4*noHidden)
	int i, items = ItemsInList(filter)
	if (strlen(filter) && CmpStr(filter,"*"))													// filter each list item ('*' just means 'include all')
		if (omit)
			for (i = 0; i < items; i++)
				traces = RemoveFromList(ListMatch(traces,StringFromList(i,filter)),traces)
			endfor
		else
			for (i = 0; i < items; i++)
				tempList += ListMatch(traces,StringFromList(i,filter))
			endfor
			traces = tempList
		endif
	endif
	
	traces = RemoveProfileTraceFromList(win, traces)
	traces = RemoveFromListWildCard(ksTraceIgnoreList, traces)
	return ItemsInList(traces)
end

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static function FetchAllTraces(string win, string &traces)
	traces = RemoveProfileTraceFromList(win, TraceNameList(win,";",1))
	traces = RemoveFromListWildCard(ksTraceIgnoreList, traces)
	return ItemsInList(traces)
end

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static function/S RemoveProfileTraceFromList(string win, string traces)
	int i, items = ItemsInList(traces)
	string remItem = ""
	for (i = 0; i < items; i++)
		string xAxis = StringByKey("XAXIS",  TraceInfo(win,StringFromList(items-1 - i,traces),0))
		if (StringMatch(xAxis,"Profilebottom") || StringMatch(xAxis,"Profileleft"))
			remItem += StringFromList(items-1 - i,traces)+";"
		endif
	endfor
	return RemoveFromList(remItem,traces)
end

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static function/S RemoveFromListWildCard(string zapListStr, string listStr)						// returns listStr, purged of any items that match an item in ZapListStr.
	string removeStr=""
	int i
	for(i = 0; i < ItemsInList(zapListStr); i++)
		removeStr += ListMatch(listStr, StringFromList(i, zapListStr))							// Wildcards okay! Case insensitive.
	endfor
	return RemoveFromList(removeStr, listStr, ";", 0)
end

//+++++++++++++++++++++++++++++++ help function for Profile Axis +++++++++++++++++++++++++++++++++

static function AdjustProfileAxisScaling()
	string gName = getTopGraph()
	if (!strlen(gName) || WhichListItem("Profileleft", AxisList(gName)) < 0)
		return 0
	endif

	string traces = ""
	int i, items = FetchAllTraces(gName,traces)
	if (items == 0)
		return 0
	endif
	variable currMax = 0, dataMin = 0, yOff
	for (i = 0; i < items; i++)
		yOff = getTraceOffset(gName, StringFromList(i,traces), 1)
		if (yOff > currMax)
			currMax = yOff
			dataMin = WaveMin(TraceNameToWaveRef(gName,StringFromList(i,traces)))
		endif
	endfor
	dataMin = numtype(dataMin)!=0 ? 0 : dataMin

	DoUpdate; GetAxis/W=$gName/Q left
	variable scale = (currMax+dataMin-V_min)/(V_Max-V_min)										// calculate the ratio between highest offset and axis height
	ModifyGraph/W=$gName axisEnab(Profileleft)={0,limit(scale, 0.1, 1)}							// prevent scaling to useless values
	return 0
end

//################################### button helper function #####################################

function TraceTool_ButtonFunctions(STRUCT WMButtonAction &s) : ButtonControl
	if (s.eventCode == 2)
		s.blockReentry=1
		TraceTool_ExecuteButtonAction(s.ctrlName)
	endif
	return 0
end

//################################################################################################

function TraceTool_ExecuteButtonAction(string which)
	string gName = getTopGraph()
	if (!strlen(gName))
		return 0
	endif
	string pSingle = "TraceOffsetPanel#SingleTab"
	string pMulti  = "TraceOffsetPanel#MultiTab"
	string trace = GetUserData("TraceOffsetPanel", "", "trace")									// get current selected trace name and cursor from panel user data
	string csr   = GetUserData("TraceOffsetPanel", "", "cursor")
	
	variable xOff = 0, yOff = 0, xScale = 0, yScale = 0, setX, setY, cursorPos
	if (strlen(trace) && strlen(csr))
		getAllTraceOffsets(gName, trace, xOff, yOff, xScale, yScale)
	endif
	
	string traces = "", axis = "X"
	int GoUp = 0, i, items = FetchAllTraces(gName,traces)
	StrSwitch(which)
		case "CsrActButton":
			if (items == 0)
				break
			endif
			for (i = 0; i < 4; i++)
				string currCsr = StringFromList(i,"A;B;C;D;")
				if (!strlen(CsrInfo($currCsr,gName)))
					trace = StringFromList(0,traces)
					getAllTraceOffsets(gName, trace, xOff, yOff, xScale, yScale)
					wave w = TraceNameToWaveRef(gName,trace)
					wave/Z w_x = XWaveRefFromTrace(gName, trace)
					
					GetAxis/W=$gName/Q $StringByKey("XAXIS",TraceInfo(gName,trace,0))
					setX = (V_min+(V_max-V_min)/2 - xOff)/xScale								// center of the current axis corrected for scaling and offset of the current trace
					cursorPos = x2pnt(w,setX)
					if(WaveExists(w_x))
						FindLevel/P/Q w_x, setX;	cursorPos = !v_flag ? V_LevelX : DimSize(w,0)/2
					endif
					cursorPos = limit(cursorPos,0,DimSize(w,0)-1)
					Cursor/W=$gName/P $currCsr, $trace, cursorPos								// set a new cursor in the center of the current axis scale if possible
					
					if (strlen(csr))															// if a cursor was already active, then switch over to the newly placed one
						CheckBox $("traceCsr"+csr)		,win=$pSingle ,value=0
						CheckBox $("traceCsr"+currCsr)	,win=$pSingle ,value=1
					endif
					FetchLatestTraceValues()
					break																		// cursor set => quit
				endif
			endfor
		break
		case "ColorButton":
			ControlInfo/W=$pMulti ColorSelect;		string Table = S_value
			ControlInfo/W=$pMulti cReverseColor;	variable rev = (GetKeyState(0) & 2^0)!=0
			ColorSpectrumFromTable((V_Value || rev), table)
		break
		case "ReverseButton":
			if (items == 0)
				break
			endif
			string trList = ""
			for (i = 0; i < items; i++)															// reverse trace list
				trList += StringFromList(items-1-i,traces) + ";"
			endfor
			wave/T w_traces = ListToTextWave(traces,";")
			Make/Free/N=(items) xOffsets = getTraceOffset(gName, w_traces[p], 0)
			Make/Free/N=(items) yOffsets = getTraceOffset(gName, w_traces[p], 1)
			
			int do2Dsort = 0;	string instanceList = GrepList(traces,"#")
			if (ItemsInList(instanceList))
				do2Dsort = (WaveDims(TraceNameToWaveRef(gName, StringFromList(0, instanceList))) == 2)	// special treatment of 2D waterfall plots
			endif
			
			if (items <= 100)																	// this even works for instances (but is limited to 100 traces)
				Execute/Q/Z "ReorderTraces/W="+gName+" _front_,{"+RemoveEnding(ReplaceString(";",trList,","),",")+"}"
			else
				if (do2Dsort)
					Abort "Error: Cannot reverse 2D data larger that 100 traces."
				endif
				for(i = 0; i < items; i++)
					ReorderTraces/W=$gName _back_, {$StringFromList(i,traces)}
				endfor
			endif
			
			for (i = 0; i < items; i++)															// reapply offsets (in reverse)
				if (do2Dsort)																	// the trace labeling will stay the same)
					ModifyGraph/W=$gName offset($w_traces[items-1-i])={xOffsets[i],yOffsets[i]}
				else
					ModifyGraph/W=$gName offset($w_traces[i])={xOffsets[i],yOffsets[i]}
				endif
			endfor
			FetchLatestTraceValues()
		break
		case "ResetYButton":
			axis = "Y"
		case "ResetXButton":
			items = FetchTraces(gName,traces)
			for (i = 0; i < items; i++)
				if (CmpStr(axis,"Y") == 0)
					ModifyGraph/W=$gName offset($StringFromList(i,traces))={,0}
				else
					ModifyGraph/W=$gName offset($StringFromList(i,traces))={0,}
				endif
			endfor
			SetVariable $("vaSet"+axis+"Offset")	,win=$pMulti	,value=_NUM:0	,userdata="0"
			SetVariable $("vaScale"+axis+"Offset")	,win=$pMulti	,value=_NUM:1	,userdata="1"
			SetVariable $("vaShift"+axis+"Offset")	,win=$pMulti	,value=_NUM:0
			AdjustProfileAxisScaling()
		break
		case "trResetYButton":
			if (strlen(trace) && strlen(csr))
				ModifyGraph/W=$gName offset($trace)={,0} ,muloffset($trace)={,0}
				SetVariable trSetYOffset	,win=$pSingle	,value=_NUM:0
				SetVariable trScaleYOffset	,win=$pSingle	,value=_NUM:1
				SetVariable trShiftYOffset	,win=$pSingle	,value=_NUM:vcsr($csr, gName)
				SetVariable trScaleNorm		,win=$pSingle	,value=_NUM:vcsr($csr, gName)
				AdjustProfileAxisScaling()
			endif
		break
		case "trResetXButton":
			if (strlen(trace) && strlen(csr))
				ModifyGraph/W=$gName offset($trace)={0,} ,muloffset($trace)={0,}
				SetVariable trSetXOffset	,win=$pSingle	,value=_NUM:0
				SetVariable trScaleXOffset	,win=$pSingle	,value=_NUM:1
				SetVariable trShiftXOffset	,win=$pSingle	,value=_NUM:hcsr($csr, gName)
				AdjustProfileAxisScaling()
			endif
		break
		case "trGoToMaxButton":
			GoUp = 1
		case "trGoToMinButton":
			if (strlen(trace) && strlen(csr))
				wave w = CsrWaveRef($csr,gName)
				WaveStats/Q/P w
				setX = GoUp ? V_maxRowLoc : V_minRowLoc
				setY = GoUp ? V_maxColLoc : V_minColLoc
				if (strlen(ImageInfo(gName,trace,0)))
					Cursor/P/W=$gName/I $csr, $trace, setX, setY								// move cursor to image position
				else
					Cursor/P/W=$gName	$csr, $trace, setX										// move cursor to trace position
				endif
				FetchLatestCursorValues(csr)
			endif
		break
		case "trNextTraceButton":
			GoUp = 1
		case "trPrevTraceButton":
			if (!strlen(trace) || !strlen(csr))
				break
			endif
			
			int currTrace = WhichListItem(trace, traces)
			string nextTrace = ""
			if (GoUp)
				if (currTrace < items-1)
					nextTrace = StringFromList(currTrace+1,traces)
				endif
			else
				if (currTrace > 0)
					nextTrace = StringFromList(currTrace-1,traces)
				endif
			endif
			if (strlen(ImageInfo(gName,trace,0)) || !strlen(nextTrace))							// already at end or image plot
				break
			endif
			
			setX = hcsr($csr, gName)*xScale + xOff												// correct for current trace's offsets
			getAllTraceOffsets(gName, nextTrace, xOff, yOff, xScale, yScale)
			setX = (setX - xOff)/xScale															// correct for next trace's offsets
			
			wave w = TraceNameToWaveRef(gName,nextTrace)
			wave/Z w_x = XWaveRefFromTrace(gName, nextTrace)
			cursorPos = x2pnt(w,setX)
			if(WaveExists(w_x))
				FindLevel/P/Q w_x, setX;	cursorPos = !v_flag ? V_LevelX : DimSize(w,0)/2
			endif
			
			Cursor/W=$gName/P $csr, $nextTrace, cursorPos										// move cursor to next trace
			FetchLatestCursorValues(csr)
		break
		case "trAddXButton":
			GoUp = 1
		case "trRemXButton":
			if (!strlen(trace) || !strlen(csr))
				break
			endif
			
			GoUp = GoUp == 0 ? -1 : GoUp
			ControlInfo/W=$pSingle trOffxDelta
			setX = V_Value
			if (setX < 0)
				ModifyGraph/W=$gName muloffset($trace)={-GoUp*abs(xScale),}
			else
				xScale = xScale == -1 ? 0 : xScale
				ModifyGraph/W=$gName muloffset($trace)={abs(xScale),}
			endif
			ModifyGraph/W=$gName offset($trace)={(GoUp*abs(setX)+xOff),}
			FetchLatestCursorValues(csr, keepDelta=1)
		break
		case "trAddYButton":
			GoUp = 1
		case "trRemYButton":
			if (!strlen(trace) || !strlen(csr))
				break
			endif
			
			GoUp = GoUp == 0 ? -1 : GoUp
			ControlInfo/W=$pSingle trOffyDelta
			setY = V_Value
			if (setY < 0)
				ModifyGraph/W=$gName muloffset($trace)={,-GoUp*abs(yScale)}
			else
				yScale = yScale == -1 ? 0 : yScale
				ModifyGraph/W=$gName muloffset($trace)={,abs(yScale)}
			endif
			ModifyGraph/W=$gName offset($trace)={,(GoUp*abs(setY)+yOff)}
			AdjustProfileAxisScaling()
			FetchLatestCursorValues(csr, keepDelta=1)
		break
		case "trNormTo1Button":
			if (strlen(trace) && strlen(csr))
				if (vcsr($csr, gName) != 0)
					ModifyGraph/W=$gName muloffset($trace)={,1/vcsr($csr, gName)}
					FetchLatestCursorValues(csr)
				endif
			endif
		break
		case "trApplyCurButton":
			if (strlen(trace))																	// falls through with only the current trace selected
				traces = trace
				items  = 1
			else
				break
			endif
		case "vaApplyAllButton":
			for (i = 0; i < items; i++)
				trace = StringFromList(i,traces)
				getAllTraceOffsets(gName, trace, xOff, yOff, xScale, yScale)
				
				wave work  = TraceNameToWaveRef(gName, trace)									// apply all scales and offsets to 1D data
				wave/Z w_x = XWaveRefFromTrace(gName, trace)
				if (WaveDims(work) != 1)														// just in case there is 2D data mixed in
					continue
				endif
				work *= yScale
				work += yOff
				if(WaveExists(w_x))
					w_x *= xScale
					w_x += xOff
				else
					SetScale/P x, DimOffset(work,0) * xScale + xOff, DimDelta(work,0) * xScale, WaveUnits(work,0), work
				endif
				ModifyGraph/W=$gName offset($trace)={0,0} ,muloffset($trace)={0,0}
				
				string lognote = ""																// write a simple log
				lognote += SelectString((xScale!=1),"","\rx-scale="+num2str(xScale))
				lognote += SelectString((xOff!=0),"","\rx-offset="+num2str(xOff))
				lognote += SelectString((yScale!=1),"","\ry-scale="+num2str(yScale))
				lognote += SelectString((yOff!=0),"","\ry-offset="+num2str(yOff))
				if (strlen(lognote))
					lognote[0] = "\rscaling and offsets applied:"
					Note work, lognote
				endif
			endfor
			FetchLatestTraceValues()
		break
	EndSwitch
	return 0
end

//#################################### popup helper function #####################################

function TraceTool_ProfileSelect(STRUCT WMPopupAction &s) : PopupMenuControl
	if (s.eventCode != 2)
		return 0
	endif
	string gName = getTopGraph()
	if (!strlen(gName))
		return 0
	endif
	
	int noSelect = StringMatch(s.popStr,"none")
	variable axFract = 0.73
	variable axRange = noSelect ? 1 : axFract-0.03
	
	SetActiveSubwindow $gName																	// make sure to do stuff in the graph window
	string axes = AxisList(gName)
	if (FindListItem("Profileleft", axes) != -1)												// see if the additional Profile axis is attached
		RemoveFromGraph/W=$gName $("AddTraceProfile")											// remove the old trace (if present)
	endif
	if (WhichListItem("bottom",axes) > -1)														// make the bottom axis shorter
		ModifyGraph/W=$gName axisEnab(bottom)={0,axRange}
	endif
	if (WhichListItem("top",axes) > -1)
		ModifyGraph/W=$gName axisEnab(top)={0,axRange}
	endif
	if (noSelect)																				// do nothing
		return 0
	endif
	
	AppendToGraph/W=$gName/C=(0,15872,65280)/B=Profilebottom/L=Profileleft/VERT $s.popStr/TN=AddTraceProfile	// append the Profile trace
	ModifyGraph/W=$gName axisEnab(Profilebottom)={axFract+0.03,1}	
	ModifyGraph/W=$gName freePos(Profileleft)   ={axFract+0.03,kwFraction}						// move the new axes to the right position
	
	ModifyGraph/W=$gName freePos(Profilebottom)=0, mirror(Profilebottom)=2
	ModifyGraph/W=$gName tkLblRot(Profileleft)=90, tlOffset(Profileleft)=-2
	ModifyGraph/W=$gName minor(Profileleft)=1, tick(Profileleft)=2, standoff=0
	ModifyGraph/W=$gName ZisZ(Profilebottom)=1,	ZisZ(Profileleft)=1
	ModifyGraph/W=$gName mirror(Profilebottom)=2,notation(Profilebottom)=1
	AdjustProfileAxisScaling()
	return 0
end

//################################################################################################

function ColorSpectrumFromTable(int rev, string table)
	string traces = "", gName = getTopGraph()
	int i, items = FetchTraces(gName,traces)
	if (items == 0)																				// may be an image
		string image = StringFromList(0,ImageNameList(gName,";"))
		if (strlen(image))
			ModifyImage/W=$gName $image ctab={,,$table,rev}
		endif
		return 0
	endif
	
	ColorTab2Wave $table;	wave M_colors
	for (i = 0; i < items; i++)
		variable row = items == 1 ? 0 : (i/(items-1))*(DimSize(M_Colors, 0)-1)
		variable red = M_Colors[row][0], green = M_Colors[row][1], blue = M_Colors[row][2]
		variable idx = rev == 0 ? i : (items-1)-i
		ModifyGraph/W=$gName/Z rgb($StringFromList(idx,traces)) = (red, green, blue)
	endfor
	KillWaves/Z M_colors
	return 0
end

//________________________________________________________________________________________________
//
//										GRAPH SIZE PANEL
//________________________________________________________________________________________________

function BuildGraphSizePanel()
	string pName = "GraphSizePanel"
	if (WinType(pName))
		DoWindow/F $pName
		return 0
	endif
	
	variable pW = 320, pH = 518
	
	variable TFTResX, TFTResY																	// get the display resolution to align window better (center of full screen)
	if (cmpstr(IgorInfo(2),"Macintosh")==0)		// Mac
		string TFTInfo = StringByKey("RECT",ReplaceString(",RECT",StringByKey("SCREEN1",IgorInfo(0),":"),";RECT"),"=")
		TFTResX = (Str2Num(StringFromList(2,TFTInfo,",")) - Str2Num(StringFromList(0,TFTInfo,",")))*(72/ScreenResolution)  
		TFTResY = (Str2Num(StringFromList(3,TFTInfo,",")) - Str2Num(StringFromList(1,TFTInfo,",")))*(72/ScreenResolution)  
	else										// Windows
		GetWindow kwFrameInner wsizeDC
		TFTResX = (v_right-v_left)
		TFTResY = (v_bottom-v_top)
	endif
	
	NewPanel/K=1/W=((TFTResX-pW)/2,(TFTResY-pH)/2,(TFTResX+pW)/2,(TFTResY+pH)/2)/N=$pName as "Set Graph Size"
	ModifyPanel /W=$pName ,fixedSize=1 ,noEdit=1
	SetWindow $pName hook(UpdateHook)=SizePanelUpdateFunc
	SetWindow $pName userdata(procVersion)=num2str(kVersion)									// save procedure version
	
	SetDrawLayer UserBack
	SetDrawEnv linethick=2,			fillpat= 0;	DrawRect 17,10,305,198							// full rectangle
	SetDrawEnv linethick=2, dash=3,	fillpat= 0;	DrawRect 55,40,265,168							// dashed rectangle
	
	SetDrawEnv linethick=2,linefgc=(16380,28400,65535),arrow=3,arrowfat=0.80;	DrawLine 58,73,257,73		// blue arrows
	SetDrawEnv linethick=2,linefgc=(16380,28400,65535),arrow=3,arrowfat=0.80;	DrawLine 210,43,210,166
	
	SetDrawEnv linethick=2,linefgc=( 8000,39320,19940),arrow=3,arrowfat=0.80;	DrawLine 21,153,300,153		// green arrows
	SetDrawEnv linethick=2,linefgc=( 8000,39320,19940),arrow=3,arrowfat=0.80;	DrawLine 110,13,110,196
	
	SetDrawEnv linethick=2,linefgc=(36870,14755,58980),arrow=3,arrowfat=0.80;	DrawLine 123,171,123,196	// purple arrows
	SetDrawEnv linethick=2,linefgc=(36870,14755,58980),arrow=3,arrowfat=0.80;	DrawLine 123,13,123,38		// vert top
	SetDrawEnv linethick=2,linefgc=(36870,14755,58980),arrow=3,arrowfat=0.80;	DrawLine 21,95,51,95		// hor left
	SetDrawEnv linethick=2,linefgc=(36870,14755,58980),arrow=3,arrowfat=0.80;	DrawLine 270,95,300,95		// hor right
	
	TitleBox TplotW		,pos={140,45}	,size={48,16}	,title=" plot W: "	,anchor=MT ,labelBack=(16380,28400,65535)
	TitleBox TplotH		,pos={188,85}	,size={45,16}	,title=" plot H: "	,anchor=MT ,labelBack=(16380,28400,65535)
	TitleBox TgraphW	,pos={82,85}	,size={55,16}	,title=" graph H: "	,anchor=MT ,labelBack=(8000,39320,19940)
	TitleBox TgraphH	,pos={135,125}	,size={58,16}	,title=" graph W: "	,anchor=MT ,labelBack=(8000,39320,19940)
	ModifyControlList ControlNameList(pName,";","T*")	,fixedSize=1	,frame=0	,fStyle=1	,fColor=(65535,65535,65535)
	
	SetVariable VLeftMargin		,pos={10,103}	,value=_NUM:0
	SetVariable VBottomMargin	,pos={133,175}	,value=_NUM:0
	SetVariable VRightMargin	,pos={250,103}	,value=_NUM:0
	SetVariable VTopMargin		,pos={133,16}	,value=_NUM:0
	SetVariable VGraphHeight	,pos={80,103}	,value=_NUM:0
	SetVariable VPlotWidth		,pos={133,63}	,value=_NUM:0
	SetVariable VGraphWidth		,pos={133,143}	,value=_NUM:0
	SetVariable VPlotHeight		,pos={180,103}	,value=_NUM:0
	SetVariable VMagnification	,pos={235,175}	,value=_NUM:0
	SetVariable VGraphFont		,pos={35,175}	,value=_NUM:0
	ModifyControlList ControlNameList(pName,";","V*")	,limits={0,2000,1}	,size={60,18} ,bodyWidth=60 ,proc=Gsize_VarControl
	
	SetVariable VMagnification	,limits={0,8,0} 	,bodyWidth=40	,title="magn.:"	,format="%g x"
	SetVariable VGraphFont		,limits={0,100,0}	,bodyWidth=40	,title="font:"							,help={"Sets the global graph font. If shift+enter is pressed instead the global axes font is changed. Insert 0 to reset to 'auto'"}
	
	Button ResetAllButton		,pos={27,16}	,size={60,20}	,title="reset all"							,help={"Resets all size controls to automatic mode."}
	TitleBox SwapTitle			,pos={147,85}	,size={60,23}	,title="\JCswap:"							,frame=0
	PopupMenu SizeUnit			,pos={210,16}	,size={85,23}	,title="unit: "
	CheckBox SwapAxes			,pos={154,105}	,size={23,23}	,title=""						,value=0	,proc=Gsize_SwapGraph
	CheckBox cPrint				,pos={18,205}	,size={23,23}	,title="Print Commands"			,value=0	,help={"Prints modify commands into history."}
	CheckBox KeepAutoMode		,pos={145,205}	,size={23,23}	,title="Keep Auto-Size Mode"	,value=1	,help={"Will reset the mode to Auto after changing the size."}
	
	GroupBox GrAxisToggle		,pos={8,228}	,size={304,92}	,title="Toggle Axis Ticks & Labels (Set Margins):"	,fStyle=1	,frame=0	,fColor=(36870,14755,58980)
	CheckBox cAxis_left			,pos={49,248}	,size={23,23}	,title="Left"					,value=0
	CheckBox cAxis_bottom		,pos={116,248}	,size={23,23}	,title="Bottom"					,value=0
	CheckBox cAxis_right		,pos={183,248}	,size={23,23}	,title="Right"					,value=0
	CheckBox cAxis_top			,pos={250,248}	,size={23,23}	,title="Top"					,value=0
	
	SetVariable tlAxis_left		,pos={49,270}	,value=_NUM:0	,title="Tick:"
	SetVariable tlAxis_bottom	,pos={116,270}	,value=_NUM:0
	SetVariable tlAxis_right	,pos={183,270}	,value=_NUM:0
	SetVariable tlAxis_top		,pos={250,270}	,value=_NUM:0
	
	SetVariable alAxis_left		,pos={49,294}	,value=_NUM:0	,title="Axis:"
	SetVariable alAxis_bottom	,pos={116,294}	,value=_NUM:0
	SetVariable alAxis_right	,pos={183,294}	,value=_NUM:0
	SetVariable alAxis_top		,pos={250,294}	,value=_NUM:0
	
	GroupBox GrSizeSetting		,pos={8,328}	,size={304,75}	,title="Set & Copy Graph Sizes:"	,fStyle=1	,frame=0	,fColor=(8000,39320,19940)
	PopupMenu AspectSelect		,pos={20,350}	,size={185,22}	,title="Aspect Ratio:"
	PopupMenu FormatSelect		,pos={20,376}	,size={185,22}	,title="Set Format:"
	Button CopyAllButton		,pos={215,348}	,size={90,22}	,title="Copy Sizes"							,help={"Copies all graph-size values and settings from the panel."}
	Button PasteAllButton		,pos={215,374}	,size={90,22}	,title="Paste to Top"						,help={"Paste copied size settings to the top graph."}
	
	GroupBox GrStyleSetting		,pos={8,410}	,size={304,101}	,title="Set & Copy Graph Styles:"	,fStyle=1	,frame=0
	PopupMenu GrStyle			,pos={20,431}	,size={285,22}	,title="Style:"
	Button ApplyButton			,pos={215,456}	,size={90,22}	,title="\f01Apply to Top"					,help={"Applies selected style to the top graph."}
	Button CopyStyleButton		,pos={18,456}	,size={55,47}	,title="Copy\rStyle"						,help={"Copies the style of the top graph (hold shift to save style to the clipboard)."}
	CheckBox CopyAxes			,pos={80,460}	,size={60,22}	,title="Include Axis Scaling"	,value=1	,help={"Copies the axis scaling as well."}
	CheckBox CopyToClipboard	,pos={80,485}	,size={60,22}	,title="Copy to Clipboard" 		,value=0	,help={"Copies the style to the clipboard."}
	Button RevertButton			,pos={215,482}	,size={90,22}	,title="Revert to Last"						,help={"Reverts to last style before Apply was pressed."}
	
	PopupMenu GrStyle			,mode=1	,value= Gsize_FetchGraphMacroList()						,bodyWidth=255
	PopupMenu SizeUnit			,mode=1	,value= #"\"points;inches;cm\""							,bodyWidth=60	,proc=Gsize_PresetPopup
	PopupMenu AspectSelect		,mode=1	,value= #("\""+kAspectSelectList+"\"")					,bodyWidth=115	,proc=Gsize_PresetPopup
	PopupMenu FormatSelect		,mode=1	,value= #("\""+kFormatSelectList+"\"")					,bodyWidth=115	,proc=Gsize_PresetPopup
	
	ModifyControlList ControlNameList(pName,";","tlAxis_*")	,limits={-50,50,1}	,size={52,18}	,bodyWidth=52	,proc=Gsize_VarControl
	ModifyControlList ControlNameList(pName,";","alAxis_*")	,limits={-50,50,1}	,size={52,18}	,bodyWidth=52	,proc=Gsize_VarControl
	ModifyControlList ControlNameList(pName,";","*Button")														,proc=Gsize_CopyPaste
	ModifyControlList ControlNameList(pName,";","cAxis_*")														,proc=Gsize_ToggleAxis
	
	if (CmpStr(IgorInfo(2), "Macintosh") == 0)
		ModifyControlList ControlNameList(pName,";","*") ,fsize=10
	endif
	
	NVAR/Z unit = root:$ksGraphSizeUnitSave														// load saved settings
	if (NVAR_Exists(unit))
		PopupMenu SizeUnit ,mode=unit
		string ctrlList = RemoveFromList("VMagnification;VGraphFont;",ControlNameList(pName,";","V*"))
		variable div  = unit == 1 ? 1 : (unit == 2 ? 72 : 72/2.54)								// in points, inches or cm
		variable step = unit == 1 ? 1 : (unit == 2 ? 0.014 : 0.036)
		ModifyControlList ctrlList ,limits={0,8000/div,step}									// official max size is 8000 points
	endif
	
	string gName = getTopMainGraph()
	if (strlen(gName))
		AutoPositionWindow/M=0/R=$gName $pName
	endif
	
	FetchLatestGraphSizeValues()
	return 0
end
//++++++++++++++++++++++++++++++++ panel update helper function ++++++++++++++++++++++++++++++++++
function SizePanelUpdateFunc(STRUCT WMWinHookStruct &s)
	if (s.eventCode == 0)
		FetchLatestGraphSizeValues()
		return 1
	endif
	return 0
end

//################################################################################################

function/S Gsize_FetchGraphMacroList()
	string list = MacroList("*",";","KIND:1,SUBTYPE:GraphStyle")
	string copy = GetUserData("GraphSizePanel", "", "CopyStyle")
	if (strlen(copy))
		list += "Copied Style"
	endif
	copy = GetUserData("GraphSizePanel", "", "CopyStyleName")
	if (strlen(copy))
		list += " ("+copy+")"
	endif
	return list + ";From Clipboard;"
end

//################################################################################################

function Gsize_SwapGraph(STRUCT WMCheckboxAction& s) : CheckBoxControl
	if (s.eventCode != 2)
		return 0
	endif
	string gName = getTopGraph()
	if (!strlen(gName))
		return 0
	endif
	
	ControlInfo/W=$(s.win) AspectSelect;	string currAspect = S_Value
//	ChangeAspectRatio("Free")																	// release the aspect mode and set again later (otherwise swap will switch this too)
	ModifyGraph/W=$gName/Z swapXY=s.checked
	ControlInfo/W=$(s.win) cPrint
	if (V_Value && V_flag == 2)
		print "•ModifyGraph swapXY="+num2str(s.checked)
	endif
//	ChangeAspectRatio(currAspect)
	FetchLatestGraphSizeValues()
	return 0
end

//################################################################################################

function Gsize_ToggleAxis(STRUCT WMCheckboxAction& s) : CheckBoxControl
	if (s.eventCode != 2)
		return 0
	endif
	string gName = getTopGraph()
	if (strlen(gName))
		SetAxisLabelsAndTicks(ReplaceString("cAxis_",s.ctrlName,""), gName, !s.checked)
		FetchLatestGraphSizeValues()
	endif
	return 0
end
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
static function SetAxisLabelsAndTicks(string axName, string gName, variable off)
	if (FindListItem(axName,AxisList(gName)) > -1)
		ModifyGraph/W=$gName tick($axName)=3*off
		ModifyGraph/W=$gName noLabel($axName)=1*off
		ModifyGraph/W=$gName margin($axName)=25*off
		ControlInfo/W=GraphSizePanel cPrint
		if (V_Value)
			print "•ModifyGraph tick("+axName+")="+num2str(3*off)+", noLabel("+axName+")="+num2str(1*off)+", margin("+axName+")="+num2str(25*off)
		endif
	endif
	return 0
end

//################################################################################################

function Gsize_CopyPaste(STRUCT WMButtonAction &s) : ButtonControl
	if (s.eventCode != 2)
		return 0
	endif
	string gName = getTopGraph()
	if (!strlen(gName))
		return 0
	endif
	string topLevelgName = StringFromList(0, gName, "#")

	string ctrlList = "", settings = "", Style = ""
	int i
	StrSwitch(s.ctrlName)
		case "ResetAllButton":
			ModifyGraph/W=$gName/Z width=0, height=0, margin=0
			ModifyGraph/W=$gName/Z gfSize=0, fSize=0, lblMargin=0, tlOffset=0
//			ModifyGraph/W=$gName/Z tick=0,noLabel=0, swapXY=0, expand=0
		break
		case "CopyAllButton":
			ctrlList = ControlNameList(s.win, ";", "*Axis_*")
			for (i = 0; i < ItemsInList(ctrlList); i++)
				ControlInfo/W=$(s.win) $StringFromList(i,ctrlList)
				settings += StringFromList(i,ctrlList) + ":" + num2str(V_Value) + ";"
			endfor
			ctrlList = RemoveFromList("VGraphWidth;VGraphHeight;VGraphFont;",ControlNameList(s.win, ";", "V*"))
			for (i = 0; i < ItemsInList(ctrlList); i++)
				ControlInfo/W=$(s.win) $StringFromList(i,ctrlList)
				settings += StringFromList(i,ctrlList) + ":" + num2str(V_Value) + ";"
			endfor
			ControlInfo/W=$(s.win) VGraphFont
			settings += "VGraphFont:" + num2str(V_Value*StringMatch(S_recreation,"*fstyle=1*")) + ";"		// font is set or release depending on its 'bold' state
			
			ControlInfo/W=$(s.win) SwapAxes;		settings +=  "SwapAxes:" + num2str(V_Value) + ";"
			ControlInfo/W=$(s.win) SizeUnit;		settings +=  "SizeUnit:" + S_Value + ";"
			ControlInfo/W=$(s.win) AspectSelect;	settings +=  "Aspect:" + S_Value + ";"
			ControlInfo/W=$(s.win) FormatSelect;	settings +=  "Format:" + S_Value + ";"
			SetWindow $(s.win) ,userdata(saveSettings) = settings
		break
		case "PasteAllButton":
			settings = GetUserData(s.win, "", "saveSettings")
			if (strlen(settings) == 0)
				break
			endif
			variable swap = str2num(StringByKey("SwapAxes", settings))
			CheckBox SwapAxes  		,win=$(s.win)	,value=swap
			PopupMenu SizeUnit 		,win=$(s.win)	,popMatch=StringByKey("SizeUnit", settings)
			PopupMenu FormatSelect	,win=$(s.win)	,popMatch=StringByKey("Format", settings)
			PopupMenu AspectSelect	,win=$(s.win)	,popMatch=StringByKey("Aspect", settings)
			
			variable cValue; string cName = ""
			ctrlList = ControlNameList(s.win, ";", "cAxis_*")
			for (i = 0; i < ItemsInList(ctrlList); i++)
				cName  = StringFromList(i,ctrlList)
				cValue = str2num(StringByKey(cName, settings)) == 0
				SetAxisLabelsAndTicks(ReplaceString("cAxis_",cName,""), gName, cValue)
			endfor
			
			ctrlList = ControlNameList(s.win, ";", "tlAxis_*")+ControlNameList(s.win, ";", "alAxis_*")
			for (i = 0; i < ItemsInList(ctrlList); i++)
				cName  = StringFromList(i,ctrlList)
				cValue = str2num(StringByKey(cName, settings))
				SetVariable $cName	,win=$(s.win) ,value=_NUM:cValue
				ControlInfo/W=$(s.win) $cName
				if (V_disable == 0)
					ExecuteVarChange(cName, cValue)
				endif
			endfor
			
			ctrlList = RemoveFromList("VGraphWidth;VGraphHeight;",ControlNameList(s.win, ";", "V*"))
			for (i = 0; i < ItemsInList(ctrlList); i++)
				cName  = StringFromList(i,ctrlList)
				cValue = str2num(StringByKey(cName, settings))
				SetVariable $cName	,win=$(s.win) ,value=_NUM:cValue
				ExecuteVarChange(cName, cValue)
			endfor
			
			ModifyGraph/W=$gName/Z swapXY=swap
			string aspect = StringByKey("Aspect", settings)
			if (!StringMatch(aspect,"Free"))													// actually apply aspect ratio setting as well (should come after swapXY)
				ChangeAspectRatio(aspect)
			endif
		break
		case "ApplyButton":																		// applies selected graph style
			ControlInfo/W=$(s.win) GrStyle;	Style = S_value
			string lastStyle = WinRecreation(topLevelgName,0)
			int noRevert = (StringMatch(lastStyle,"*TransformMirror*") || StringMatch(lastStyle,"*AxisTransform*"))	// graphs with a transformed axis cannot be recreated
			if (noRevert)
				DoAlert 1, "Reverting the graph will not be possible. Apply anyway?"
				if (V_flag != 1)
					break
				endif
			endif
			
			int doStyle = StringMatch(Style,"Copied Style*") ? 1 : (StringMatch(Style,"From Clipboard*") ? 2 : 0)
			
			ControlInfo/W=$(s.win) SwapAxes;	int doSwap = V_Value
			ControlInfo/W=$(s.win) cPrint;		int doPrint = (V_Value && V_flag == 2)

			Switch (doStyle)
				case 1:	// copied style
					Style = GetUserData(s.win, "", "CopyStyle")
				break
				case 2:	// clipboard style
					Style = GetScrapText()
					if (!StringMatch(Style,"*ModifyGraph*"))
						Abort "It seems there are no valid style commands in the clipboard.\r\rTry to first copy a style (e.g., by using Copy Style while holding Shift)."
					endif
					Style = ReplaceString("\r", SpectraGraphTools#Gsize_cleanupMacro(Style), ";")
				break
			EndSwitch
			
			DoWindow/F $topLevelgName															// make sure the window is in front for the macro to work on
			if (strlen(Style))
				if (doStyle)
					if (StringMatch(Style,"swapXY=1") != doSwap)								// if the swap states between the copied style and the target graphs do not match ...
						if (doSwap)
							Style = ReplaceString("height={Aspect",Style,"width={Aspect")		// ... exchange the applied direction
						else
							Style = ReplaceString("width={Aspect",Style,"height={Aspect")
						endif
					endif
					for (i = 0; i < ItemsInList(Style); i++)
						Execute/Q/Z StringFromList(i,Style)
						if (doPrint && V_flag == 0)
							print "•"+StringFromList(i,Style)
						endif
					endfor
				else
					Execute/Q Style+"()"
					if (doPrint)
						print "•"+Style+"()"
					endif
				endif
			endif
			DoWindow/F $(s.win)
			
			if (!noRevert)
				SetWindow $(s.win), userdata(LastgName)=topLevelgName							// save into panel's user data
				SetWindow $(s.win), userdata(LastStyle)=lastStyle
			endif
		break
		case "RevertButton":																	// copies top-graph's style
			Style = GetUserData(s.win, "", "LastStyle")											// check for existence of backup style
			if (strlen(Style) > 0)
				KillWindow $topLevelgName
				string/G ExeCmd = Style
				Execute/P/Q "Execute ExeCmd"
				Execute/P/Q "KillStrings/Z ExeCmd"
			endif
		break
		case "CopyStyleButton":																	// copies style of selected graph
			Style = WinRecreation(gName,1)
			string childList = ChildWindowList(gName), childStyle = ""							// need to clean the style macro of any subwindow styles (won't work anyway)
			for (i = 0; i < ItemsInList(childList); i++)
				childStyle = WinRecreation(gName+"#"+StringFromList(i,childList),1)
				childStyle = RemoveListItem(0,childStyle, "\r")									// remove "proc" line
				childStyle = RemoveListItem(0,childStyle, "\r")									// remove "pauseupdate" line
				childStyle = RemoveListItem(ItemsInList(childStyle,"\r") - 1,childStyle, "\r")	// remove "endmacro" line
				Style = ReplaceString(childStyle,Style,"")										// remove child window style
			endfor
			Style = SpectraGraphTools#Gsize_cleanupMacro(Style)
			ControlInfo/W=$(s.win) CopyAxes														// include axes if checkbox is active
			if (!V_Value)
				Style = GrepList(Style, "^(SetAxis)", 1, "\r")
			endif
			ControlInfo/W=$(s.win) CopyToClipboard	
			if ((GetKeyState(0) & 2^2) || (V_flag == 2 && V_value))								// shift key saves to clipboard
				PutScrapText Style
				PopupMenu GrStyle ,win=$(s.win) ,popMatch="From Clipboard"
			else
				Style = ReplaceString("\r", Style, ";")
				SetWindow $(s.win), userdata(CopyStyle)=Style									// save into panel's user data
				SetWindow $(s.win), userdata(CopyStyleName)=gName
				PopupMenu GrStyle ,win=$(s.win) ,popMatch="Copied Style*"
			endif
		break
	EndSwitch
	FetchLatestGraphSizeValues()
	return 0
end

//################################################################################################

static function/S Gsize_cleanupMacro(string str)
	str = RemoveFromList(ListMatch(str,"*Proc*","\r"), str, "\r")								// remove "proc" line
	str = RemoveFromList(ListMatch(str,"*PauseUpdate*","\r"), str, "\r")						// remove "pauseupdate" line
	str = RemoveFromList(ListMatch(str,"*EndMacro*","\r"), str, "\r")							// remove "endmacro" line
	str = ReplaceString("\t", str, "")
	str = GrepList(str, "^(DefineGuide|ControlBar|SetWindow|Cursor|ListBox|CheckBox|PopupMenu|ValDisplay|SetVariable|Button|TitleBox|Slider|GroupBox|NewPanel|ModifyPanel|RenameWindow|SetActiveSubwindow|ShowTools)", 1, "\r") // remove control items
	str = GrepList(str, "^(ModifyGraph/Z rgb|ModifyGraph/Z muloffset|ModifyGraph/Z offset|ShowInfo)", 1, "\r")
	return str
end

//################################################################################################

function Gsize_VarControl(struct WMSetVariableAction &s) : SetVariableControl
	if (s.eventCode == 1 || s.eventCode == 2)
		ExecuteVarChange(s.ctrlName, s.dval)
	endif
	return 0
end
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
static function ExecuteVarChange(string ctrlName, variable val)
	string gName = getTopGraph()
	string pName = "GraphSizePanel"
	if (!strlen(gName))
		return 0
	endif
	
	ControlInfo/W=$pName SizeUnit; variable modifier = V_Value == 1 ? 1 : (V_Value == 2 ? 72 : 72/2.54)		// in points, inches or cm
	ControlInfo/W=$pName KeepAutoMode;	int keepAuto = V_Value												// reset to Auto scale mode after size change
	ControlInfo/W=$pName cPrint;		int doPrint  = V_flag == 2 ? V_Value :0
	
	string cmd = ""
	StrSwitch(ctrlName)
		case "VLeftMargin":
			ModifyGraph/W=$gName margin(left) = val*modifier
			cmd = "margin(left)="+num2str(val*modifier)
		break
		case "VBottomMargin":
			ModifyGraph/W=$gName margin(bottom) = val*modifier
			cmd = "margin(bottom)="+num2str(val*modifier)
		break
		case "VRightMargin":
			ModifyGraph/W=$gName margin(right) = val*modifier
			cmd = "margin(right)="+num2str(val*modifier)
		break
		case "VTopMargin":
			ModifyGraph/W=$gName margin(top) = val*modifier
			cmd = "margin(top)="+num2str(val*modifier)
		break
		case "VGraphHeight":
			ControlInfo/W=$pName VTopMargin;	variable mTop = V_Value
			ControlInfo/W=$pName VBottomMargin;	variable mBot = V_Value
			variable pHeight = val-mTop-mBot
			if (val == 0)
				ModifyGraph/W=$gName margin(top) = 0, margin(bottom) = 0,  height = 0
				cmd = "margin(top)=0, margin(bottom)=0,  height=0"
			elseif (pHeight > 0)
				ModifyGraph/W=$gName height = pHeight*modifier
				cmd = "height="+num2str(pHeight*modifier)
			endif
		break
		case "VGraphWidth":
			ControlInfo/W=$pName VLeftMargin;	variable mLeft = V_Value
			ControlInfo/W=$pName VRightMargin;	variable mRight = V_Value
			variable pWidth = val-mLeft-mRight
			if (val == 0)
				ModifyGraph/W=$gName margin(left) = 0, margin(right) = 0,  width = 0
				cmd = "margin(left)=0, margin(right)=0,  width=0"
			elseif (pWidth > 0)
				ModifyGraph/W=$gName width = pWidth*modifier
				cmd = "width="+num2str(pWidth*modifier)
			endif
		break
		case "VPlotHeight":
			ModifyGraph/W=$gName height = val*modifier
			cmd = "height="+num2str(val*modifier)
		break
		case "VPlotWidth":
			ModifyGraph/W=$gName width = val*modifier
			cmd = "width="+num2str(val*modifier)
		break
		case "VMagnification":
			variable Mag = val == 1 ? 0 : val	// reset to none
			ModifyGraph/W=$gName expand = Mag
			cmd = "expand="+num2str(Mag)
		break
		case "VGraphFont":
			if ((GetKeyState(0) & 2^2) != 0)	// shift pressed
				ModifyGraph/W=$gName fSize = val
				cmd = "fSize="+num2str(val)
			else
				ModifyGraph/W=$gName gfSize = val
				cmd = "gfSize="+num2str(val)
			endif
		break
		case "tlAxis_left":
			ModifyGraph/W=$gName tlOffset(left) = val
			cmd = "tlOffset(left)="+num2str(val)
		break
		case "tlAxis_bottom":
			ModifyGraph/W=$gName tlOffset(bottom) = val
			cmd = "tlOffset(bottom)="+num2str(val)
		break
		case "tlAxis_right":
			ModifyGraph/W=$gName tlOffset(right) = val
			cmd = "tlOffset(right)="+num2str(val)
		break
		case "tlAxis_top":
			ModifyGraph/W=$gName tlOffset(top) = val
			cmd = "tlOffset(top)="+num2str(val)
		break
		case "alAxis_left":
			ModifyGraph/W=$gName lblMargin(left) = val
			cmd = "lblMargin(left)="+num2str(val)
		break
		case "alAxis_bottom":
			ModifyGraph/W=$gName lblMargin(bottom) = val
			cmd = "lblMargin(bottom)="+num2str(val)
		break
		case "alAxis_right":
			ModifyGraph/W=$gName lblMargin(right) = val
			cmd = "lblMargin(right)="+num2str(val)
		break
		case "alAxis_top":
			ModifyGraph/W=$gName lblMargin(top) = val
			cmd = "lblMargin(top)="+num2str(val)
		break
	EndSwitch
	
	if (doPrint)
		print "•ModifyGraph "+cmd
	endif
	
	if (keepAuto)
		DoUpdate/W=$gName
		if (StringMatch(ctrlName,"*Height"))
			ModifyGraph/W=$gName height = 0
		elseif (StringMatch(ctrlName,"*Width"))
			ModifyGraph/W=$gName width = 0
		endif
	endif
	FetchLatestGraphSizeValues()
	return 0
end

//################################################################################################

function Gsize_PresetPopup(STRUCT WMPopupAction &s) : PopupMenuControl
	if (s.eventCode == 2)
		StrSwitch(s.ctrlName)
			case "AspectSelect":
				ChangeAspectRatio(s.popStr)
			break
			case "FormatSelect":
				ChangeGraphFormat(s.popStr)
			break
			case "SizeUnit":
				string ctrlList = RemoveFromList("VMagnification;VGraphFont;",ControlNameList(s.win,";","V*"))
				variable div  = s.popNum == 1 ? 1 : (s.popNum == 2 ? 72 : 72/2.54)				// in points, inches or cm
				variable step = s.popNum == 1 ? 1 : (s.popNum == 2 ? 0.014 : 0.036)
				ModifyControlList ctrlList ,win=$(s.win) ,limits={0,8000/div,step*2}			// official max size is 8000 points
				variable/G root:$ksGraphSizeUnitSave = s.popNum									// save settings into global variable
				FetchLatestGraphSizeValues()
			break
		EndSwitch
	endif
	return 0
end

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static StrConstant kAspectSelectList = "Free;1:1;-;4:3;3:2;16:9;-;2:3;3:4;4:5"

static function ChangeAspectRatio(string which)
	string gName = getTopGraph()
	if (!strlen(gName))
		return 0
	endif
	ControlInfo/W=GraphSizePanel SwapAxes;	int isSwapped = V_Value
	ControlInfo/W=GraphSizePanel cPrint;	int doPrint = (V_Value && V_flag == 2)	
	string mode = ""
	StrSwitch (which)
		case "Free":
			mode = "0"
		break
		case "1:1":
			mode = "{Aspect,1}"
		break
		case "4:3":
			mode = "{Aspect,3/4}"
		break
		case "3:2":
			mode = "{Aspect,2/3}"
		break
		case "16:9":
			mode = "{Aspect,9/16}"
		break
		case "2:3":
			mode = "{Aspect,3/2}"
		break
		case "3:4":
			mode = "{Aspect,4/3}"
		break
		case "4:5":
			mode = "{Aspect,5/4}"
		break
	EndSwitch
	if (!strlen(mode))
		return 0
	endif
	
	string inactiveAxis = SelectString(isSwapped,"width","height")
	string activeAxis   = SelectString(isSwapped,"height","width")
	if (StringMatch(WinRecreation(gName,0), "*width={Aspect*"))									// make sure to not apply aspect settings in both axes => swapping has no influence on the direction, it is always width
		Execute/Q "ModifyGraph/Z/W="+gName+" "+inactiveAxis+"=0"
		if (doPrint)
			print "•ModifyGraph "+inactiveAxis+"=0"
		endif
	endif
	Execute/Q "ModifyGraph/Z/W="+gName+" "+activeAxis+"="+mode+";FetchLatestGraphSizeValues();"
	if (doPrint)
		print "•ModifyGraph "+activeAxis+"="+mode
	endif

	return 0
end

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static StrConstant kFormatSelectList = "Free Size;Paper Half;Paper Full;PPT 16:9 Half;PPT 16:9 Full;PPT 4:3 Half;PPT 4:3 Full;"

static function ChangeGraphFormat(string which)
	string gName = getTopGraph()
	if (!strlen(gName))
		return 0
	endif
	PopupMenu AspectSelect	,win=GraphSizePanel	,popMatch="Free"								// reset aspect popup
	if (StringMatch(which,"Free Size"))
		ModifyGraph/W=$gName/Z width=0, height=0, margin=0
		ModifyGraph/W=$gName/Z gfSize=0, fSize=0
		ModifyGraph/W=$gName/Z expand=0, lsize=1
		FetchLatestGraphSizeValues()
		return 0
	endif
	
	ControlInfo/W=GraphSizePanel KeepAutoMode;	int keepAuto = V_Value
	GetAxis/W=$gName/Q bottom;	int bottomOn = !V_Flag											// check which axes are active (increases the margin here)
	GetAxis/W=$gName/Q left;	int leftOn = !V_Flag
	GetAxis/W=$gName/Q right;	int rightOn = !V_Flag
	GetAxis/W=$gName/Q top;		int topOn = !V_Flag
	
	if (StringMatch(which, "Paper*")) 
		ModifyGraph/W=$gName/Z gfSize=8, fSize=8
		ModifyGraph/W=$gName/Z expand=1.25
		ModifyGraph/W=$gName/Z lsize=1.5
		
		ModifyGraph/W=$gName/Z margin(left)		= (11 + 22*leftOn	+ 10*(bottomOn || topOn && !leftOn))
		ModifyGraph/W=$gName/Z margin(bottom)	= (11 + 22*bottomOn)
		ModifyGraph/W=$gName/Z margin(top)		= (11 + 22*topOn)
		ModifyGraph/W=$gName/Z margin(right)	= (11 + 22*rightOn	+ 10*(bottomOn || topOn && !rightOn))
		
		if (StringMatch(which, "*Half"))
			ModifyGraph/W=$gName/Z width =(180 - 12*(leftOn && rightOn))
			ModifyGraph/W=$gName/Z height=(172 - 22*(bottomOn && topOn))
		elseif (StringMatch(which, "*Full"))
			ModifyGraph/W=$gName/Z width =(450 - 12*(leftOn && rightOn))
			ModifyGraph/W=$gName/Z height=(208 - 22*(bottomOn && topOn))
		endif
	endif
	
	if (StringMatch(which, "PPT*")) 
		ModifyGraph/W=$gName/Z gfSize=16, fSize=12
		ModifyGraph/W=$gName/Z expand=0.75
		ModifyGraph/W=$gName/Z lsize=1.5

		ModifyGraph/W=$gName/Z margin(left)		= (18 + 37*leftOn	+ 12*(bottomOn || topOn && !leftOn))
		ModifyGraph/W=$gName/Z margin(bottom)	= (18 + 37*bottomOn)
		ModifyGraph/W=$gName/Z margin(top)		= (18 + 37*topOn)
		ModifyGraph/W=$gName/Z margin(right)	= (18 + 37*rightOn	+ 12*(bottomOn || topOn && !rightOn))
		
		if (StringMatch(which, "*16:9*"))
			ModifyGraph/W=$gName/Z height=(360 - 37*(bottomOn && topOn))
			if (StringMatch(which, "*Half"))
				ModifyGraph/W=$gName/Z width=(380 - 25*(leftOn && rightOn))
			elseif (StringMatch(which, "*Full"))
				ModifyGraph/W=$gName/Z width=(800 - 25*(leftOn && rightOn))
			endif
		elseif (StringMatch(which, "*4:3*"))
			ModifyGraph/W=$gName/Z height=(350 - 37*(bottomOn && topOn))
			if (StringMatch(which, "*Half"))
				ModifyGraph/W=$gName/Z width=(290 - 25*(leftOn && rightOn))
			elseif (StringMatch(which, "*Full"))
				ModifyGraph/W=$gName/Z width=(600 - 25*(leftOn && rightOn))
			endif
		endif
	endif

	DoUpdate
	if (keepAuto)
		ModifyGraph/W=$gName/Z height = 0
		ModifyGraph/W=$gName/Z width = 0
	endif
	FetchLatestGraphSizeValues()
	return 0
end

//################################################################################################

function FetchLatestGraphSizeValues()
	string gName = getTopGraph(), gRec = ""
	string pName = "GraphSizePanel"
	if (!strlen(gName))
		DoWindow /T $pName, "Set Graph Size: No Graph"
		return 0
	endif
	
	DoUpdate/W=$gName
	GetWindow $gName gsize																		// values from GetWindow (sometimes off by a fraction?)
	//print V_left, V_right, V_top, V_bottom		// debug
  	variable gLeft	= V_left
  	variable gRight	= V_right
  	variable gTop	= V_top
	variable gBottom= V_bottom
  	variable gWidth	= V_right-V_left	
  	variable gHeight= V_bottom-V_top	
	GetWindow $gName psize
	//print V_left, V_right, V_top, V_bottom		// debug
	variable mLeft	= round(V_left - gLeft)
	variable mTop	= round(V_top - gTop)
	variable mRight	= round(gRight - V_right)
	variable mBottom= round(gBottom - V_bottom)
	variable pWidth	= round(V_right - V_left)
	variable pHeight= round(V_bottom - V_top)
	// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	gRec = GrepList(WinRecreation(gName, 0), "^\t(ModifyGraph)", 0 , "\r")						// if available, replace with values from recreation macro (these are exact)
	gRec = GrepList(gRec, "(?i)(height|width|margin|gfsize|swapXY)", 0 , "\r")					// extract only relevant keys
	gRec = ReplaceString("\tModifyGraph ",gRec,"")
	gRec = ReplaceString("\r",gRec,",")
	
	GetWindow $gName expand
	variable gMagn = V_Value == 0 ? 1 : V_Value													// magnification parameter
	variable gSwap = str2num(StringByKey("swapXY", gRec, "=", ","));	gSwap = numtype(gSwap) == 0 ? gSwap : 0			// swap axes parameter
	variable Rec_mLeft	 = gSwap ? str2num(StringByKey("margin(bottom)", gRec, "=", ",")): str2num(StringByKey("margin(left)", gRec, "=", ","))
	variable Rec_mTop	 = gSwap ? str2num(StringByKey("margin(right)", gRec, "=", ",")) : str2num(StringByKey("margin(top)", gRec, "=", ","))
	variable Rec_mRight	 = gSwap ? str2num(StringByKey("margin(top)", gRec, "=", ","))	 : str2num(StringByKey("margin(right)", gRec, "=", ","))
	variable Rec_mBottom = gSwap ? str2num(StringByKey("margin(left)", gRec, "=", ","))	 : str2num(StringByKey("margin(bottom)", gRec, "=", ","))
	variable Rec_pWidth	 = gSwap ? str2num(StringByKey("height", gRec, "=", ","))		 : str2num(StringByKey("width", gRec, "=", ","))
	variable Rec_pHeight = gSwap ? str2num(StringByKey("width", gRec, "=", ","))		 : str2num(StringByKey("height", gRec, "=", ","))
	variable Rec_gHeight = Rec_pHeight + Rec_mBottom + Rec_mTop
	variable Rec_gWidth	 = Rec_pWidth  + Rec_mRight  + Rec_mLeft
	variable Rec_gFont	 = str2num(StringByKey("gfSize", gRec, "=", ","))
	variable gFont = numtype(Rec_gFont) == 0 ? Rec_gFont : 10
	// ++++++++++++++++++++++++++++++++ get aspect ratio info ++++++++++++++++++++++++++++++++++++
	string whichAspect = "height="																// SelectString(gSwap, "height=", "width=") => swapping has no influence on the direction, it is always height
	variable setAspect = NaN
	int i, strpos = strsearch(gRec,whichAspect,0)
	if (strpos > -1)
		string aspectStr = gRec[strpos+strlen(whichAspect),inf]									// extract aspect ratio from graph
		strpos = strsearch(aspectStr,"}",0); aspectStr = aspectStr[0,strpos]
		sscanf aspectStr, "{Aspect,%f}", setAspect
		if (numtype(setAspect) == 0)
			variable num1, num2
			for (i = 0; i < ItemsInList(kAspectSelectList); i++)								// compare with aspect presets
				sscanf StringFromList(i,kAspectSelectList), "%f:%f", num1, num2;	num1 = num2/num1
				if (numtype(num1) == 0 && abs(setAspect-num1) < 10^-4)
					setAspect = i+1
					break
				endif
			endfor
		endif
	endif
	setAspect = numtype(setAspect) != 0 ? 1 : setAspect
	// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	mLeft	= numtype(Rec_mLeft)==0 ?	Rec_mLeft : mLeft										// if the recreation values are valid then use these
	mBottom	= numtype(Rec_mBottom)==0 ?	Rec_mBottom : mBottom
	mRight	= numtype(Rec_mRight)==0 ?	Rec_mRight : mRight
	mTop	= numtype(Rec_mTop)==0 ?	Rec_mTop : mTop
	pWidth	= numtype(Rec_pWidth)==0 ?	Rec_pWidth : pWidth
	pHeight	= numtype(Rec_pHeight)==0 ?	Rec_pHeight : pHeight
	gHeight	= numtype(Rec_gHeight)==0 ?	Rec_gHeight : pHeight + mTop + mBottom
	gWidth	= numtype(Rec_gWidth)==0 ?	Rec_gWidth : pWidth + mLeft + mRight
	
	ControlInfo/W=$pName SizeUnit;
	variable modifier = V_Value == 1 ? 1 : (V_Value == 2 ? 72 : 72/2.54)						// in points, inches or cm
	if (modifier != 1)
		mLeft	= round((mLeft	 /modifier)*1000)/1000
		mBottom	= round((mBottom /modifier)*1000)/1000
		mRight	= round((mRight	 /modifier)*1000)/1000
		mTop	= round((mTop	 /modifier)*1000)/1000
		gHeight	= round((gHeight /modifier)*1000)/1000
		pWidth	= round((pWidth	 /modifier)*1000)/1000
		gWidth	= round((gWidth	 /modifier)*1000)/1000
		pHeight	= round((pHeight /modifier)*1000)/1000
	endif
	// +++++++++++++++++++++++++++++++ apply info to panel controls ++++++++++++++++++++++++++++++
	string undoGName = GetUserData(pName, "", "LastgName")										// check for existence of a undo style for the current graph
	DoWindow /T $pName, "Set Graph Size: " + gName
	CheckBox SwapAxes			,win=$pName	,value=gSwap
	SetVariable VLeftMargin		,win=$pName	,value=_NUM:mLeft	,fstyle=(numtype(Rec_mLeft)==0)
	SetVariable VBottomMargin	,win=$pName	,value=_NUM:mBottom	,fstyle=(numtype(Rec_mBottom)==0)
	SetVariable VRightMargin	,win=$pName	,value=_NUM:mRight	,fstyle=(numtype(Rec_mRight)==0)
	SetVariable VTopMargin		,win=$pName	,value=_NUM:mTop	,fstyle=(numtype(Rec_mTop)==0)
	SetVariable VGraphHeight	,win=$pName	,value=_NUM:gHeight	,fstyle=(numtype(Rec_gHeight)==0)
	SetVariable VPlotWidth		,win=$pName	,value=_NUM:pWidth	,fstyle=(numtype(Rec_pWidth)==0)
	SetVariable VGraphWidth		,win=$pName	,value=_NUM:gWidth	,fstyle=(numtype(Rec_gWidth)==0)
	SetVariable VPlotHeight		,win=$pName	,value=_NUM:pHeight	,fstyle=(numtype(Rec_pHeight)==0)
	SetVariable VGraphFont		,win=$pName	,value=_NUM:gFont	,fstyle=(numtype(Rec_gFont)==0)
	SetVariable VMagnification	,win=$pName	,value=_NUM:gMagn
	PopupMenu AspectSelect		,win=$pName	,mode=setAspect
	Button RevertButton			,win=$pName	,disable=2*(!StringMatch(gName,undoGName))			// disable button if undo backup does not match to current graph
	
	string axes = "left;bottom;right;top;"
	for (i = 0; i < ItemsInList(axes); i++)														// set axis check-boxes
		variable isOn = 0, tickOffset = 0, lblOffset = 0
		string currAxis = StringFromList(i,axes)
		gRec = AxisInfo(gName,currAxis)
		if (strlen(gRec))
			isOn = (str2num(StringByKey("tick(x)",gRec,"=",";"))!=3 && str2num(StringByKey("noLabel(x)",gRec,"=",";"))!= 2)		// inactive if both tick(x)=3 and noLabel(x)=2
			tickOffset = str2num(StringByKey("tlOffset(x)",gRec,"=",";"))
			lblOffset  = str2num(StringByKey("lblMargin(x)",gRec,"=",";"))
		endif
		CheckBox $("cAxis_"+currAxis) ,win=$pName ,value=isOn ,disable=2*(strlen(gRec)==0)
		ControlInfo/W=$pName $("tlAxis_"+currAxis)
		if (abs(V_Flag) == 5)
			SetVariable $("tlAxis_"+currAxis) ,win=$pName ,value=_NUM:tickOffset ,disable=2*(strlen(gRec)==0)
		endif
		ControlInfo/W=$pName $("alAxis_"+currAxis)
		if (abs(V_Flag) == 5)
			SetVariable $("alAxis_"+currAxis) ,win=$pName ,value=_NUM:lblOffset ,disable=2*(strlen(gRec)==0)
		endif
	endfor
	return 0
end

//________________________________________________________________________________________________
//	attaches a quick-scaling menu to the top graph
//	version 4 -  2022/03
//________________________________________________________________________________________________

function AttachQuickScalePanel()
	string win=getTopMainGraph()
	if (strlen(win) == 0)
		return 0
	endif
	GetWindow $win hook(MainWindow)
	if (StringMatch(S_Value, "QuickScale_GraphHook"))
		return 0
	endif
	SetWindow $win hook(MainWindow)=QuickScale_GraphHook, hookevents = 4
	ShowInfo
	
	//++++++++++++++++++++++++++++++ build the panel ++++++++++++++++++++++++++++++++	
	string pPath = win+"#QuickScalePanel"
	NewPanel/HOST=$win/N=QuickScalePanel/EXT=0/W=(0,0,108,452) as "Norm & Scale"
	TitleBox tCursor		,pos={10,10}	,size={90,23}	,title="No Cursor"			,frame=0 	,fstyle=1	,userdata="none"
	Button bAddCursor		,pos={78,8}		,size={20,20}	,title="+"					,help={"Adds a new cursor to the graph."}
	Button bTracePrev		,pos={8,33}		,size={40,20}	,title="Prev"				,help={"Set cursor to previous trace in list."}
	Button bTraceNext		,pos={58,33}	,size={40,20}	,title="Next"				,help={"Set cursor to next trace in list."}
	TitleBox tFindPos		,pos={10,60}	,size={90,23}	,title="Place Cursor at:"	,frame=0
	Button bFindMin			,pos={8,80}		,size={40,20}	,title="Min"				,help={"Find the minimum data value."}
	Button bFindMax			,pos={58,80}	,size={40,20}	,title="Max"				,help={"Find the maximum data value."}
	DrawLine 8, 110, 98 ,110
	TitleBox tNormalize		,pos={10,118}	,size={90,23}	,title="Normalize to:"		,frame=0
	SetVariable vNormalize	,pos={8,138}	,size={90,23}	,title=""					,help={"Normalizes the spectrum to this value at cursor position."}
	Button bNormBack		,pos={8,163}	,size={90,20}	,title="Revert Norm."		,help={"Undoes the previous normalization."}
	DrawLine 8, 192, 98 ,192
	CheckBox cImageYShift	,pos={52,200}	,size={35,23}	,title="do Y:"				,fColor=(0,0,65280)
	TitleBox tShift			,pos={10,200}	,size={90,23}	,title="X pos:"				,frame=0
	SetVariable vShift		,pos={8,220}	,size={90,23}	,title=""					,help={"Offsets the selected axis so that the cursor will be at this value."}
	TitleBox tRelShift		,pos={10,245}	,size={90,23}	,title="Relative X shift:"	,frame=0	
	SetVariable vRelShift	,pos={8,265}	,size={90,23}	,title=""					,help={"Offsets the selected axis by this relative value."}
	DrawLine 8, 294, 98 ,294
	TitleBox tPhoton		,pos={10,300}	,size={90,23}	,title="Photon Energy"		,frame=0
	SetVariable vPhoton		,pos={8,320}	,size={90,23}	,title=""					,help={"Set the photon energy."}
	Button bSwitch			,pos={8,345}	,size={90,20}	,title="KE <> BE"			,help={"Switch between kinetic and binding energy using the set photon energy."}
	DrawLine 8, 374, 98 ,374
	TitleBox tArithmetic	,pos={10,380}	,size={90,23}	,title="Arithmetic Op.:"	,frame=0
	SetVariable vArithmetic	,pos={8,400}	,size={90,23}	,title=""					,help={"Multiply, divide, add, subtract a value by/to/from the data by this command just like in the command line."}
	CheckBox cImageOneCol	,pos={9,423}	,size={90,23}	,title="One Column"			,help={"Applies arithmetic operation to the selected column only instead of the full image."}	,value=1
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	
	SetVariable vShift		,value=_NUM:0	,limits={-inf,inf,1}	,format="%g"
	SetVariable vNormalize	,value=_NUM:0	,limits={-inf,inf,1}					,userdata(reset)="0"	,userdata(wavname)="none"
	SetVariable vRelShift	,value=_NUM:0	,limits={-inf,inf,1}	,format="%g"	,userdata(offset)="0"	,userdata(wavname)="none"
	SetVariable vPhoton		,value=_NUM:0	,limits={0,inf,0}		,format="%g"
	SetVariable vArithmetic	,value=_STR:""
	CheckBox cImageYShift	,side=1			,disable=1		,proc=QuickRescale_CheckExecute
	ModifyControlList ControlNameList(pPath,";","b*")		,proc=QuickRescale_ButtonExecute
	ModifyControlList ControlNameList(pPath,";","v*")		,proc=QuickRescale_VariableExecute
	ModifyControlList ControlNameList(pPath,";","!bAddCursor") disable = 2						// disable everything for now (other than the add cursor button)
	
	if (CmpStr(IgorInfo(2), "Macintosh") == 0)
		ModifyControlList ControlNameList(pPath,";","*")	,win=$pPath	,fsize=10
	endif
	
	SetWindow $(pPath) hook(SubWindow)=QuickScale_PanelHook										// set an additional hook to handle close events
	SetActiveSubwindow $win
	
	int i
	for (i = 0; i < ItemsInList(kUsedCursorList); i++)											// see if a cursor is set already
		string csr = StringFromList(i,kUsedCursorList)
		if (strlen(CsrInfo($csr,win)))
			string trace = StringByKey("TNAME",CsrInfo($csr,win))
			if (strlen(ImageInfo(win,trace,0)))													// update the cursor (recalculates the control values)
				Cursor/I/P $csr, $trace, pcsr($csr,win), qcsr($csr,win)
			else
				Cursor/P   $csr, $trace, pcsr($csr,win)
			endif
			break
		endif
	endfor
	
	return 0
end

//################################################################################################

function QuickScale_GraphHook(STRUCT WMWinHookStruct &s)
	string pPath = s.WinName+"#QuickScalePanel"
	int i, HookTakeover = 0
	Switch(s.eventCode)
		case 8:		// modified => if trace gets added or removed
		case 7:		// cursor moved
			if (WinType(pPath) != 7)															// something went wrong, since there is no panel => remove hook
				SetWindow $s.WinName hook(MainWindow)=$""
				return 0
			endif
			string csr = s.CursorName, trace = s.TraceName
			if (!strLen(trace))																	// if the trace is not assigned, look for other cursors
				for (i = 0; i < ItemsInList(kUsedCursorList); i++)
					csr = StringFromList(i,kUsedCursorList)
					if (strlen(CsrInfo($csr,s.WinName)))
						trace = StringByKey("TNAME",CsrInfo($csr,s.WinName))
						break
					endif
				endfor
			endif
			
			if (!strLen(trace))																			// still not trace? => abort
				ModifyControlList ControlNameList(pPath,";","!bAddCursor") ,win=$pPath ,disable = 2		// disable everything other than the add cursor button
				TitleBox tCursor		,win=$pPath ,title="No Cursor" ,userdata="none"
				CheckBox cImageYShift	,win=$pPath	,disable=1
				return 0
			endif
			HookTakeover = 1
			
			wave w = CsrWaveRef($csr,s.WinName)
			wave/Z w_x = CsrXWaveRef($csr,s.WinName)
			string wName = NameOfWave(w)
			string imgInfo = ImageInfo(s.WinName,s.TraceName,0)
			int isImg = strlen(imgInfo)

			ModifyControlList ControlNameList(pPath,";","*")		,win=$pPath ,disable = 0			// re-enable everything
			ModifyControlList ControlNameList(pPath,";","bTrace*")	,win=$pPath ,disable = 2*isImg		// disable trace controls for an image
			
			string hUnit, wOffsets
			variable hv = abs(TraceTool_FetchXShift(w, hUnit));	hv = numtype(hv) ? 0 : hv
			variable xOff = DimOffset(w,0), yOff = DimOffset(w,1), xDelta = DimDelta(w,0), yDelta = DimDelta(w,1)
			if (isImg)
				wave/Z imgXW = $(StringByKey("XWAVEDF",imgInfo)+StringByKey("XWAVE",imgInfo))
				wave/Z imgYW = $(StringByKey("YWAVEDF",imgInfo)+StringByKey("YWAVE",imgInfo))
				if (WaveExists(imgXW))
					xOff   = imgXW[0]
					xDelta = (imgXW[DimSize(imgXW,0)-1] - imgXW[0])/(DimSize(imgXW,0)-1)
				endif
				if (WaveExists(imgYW))
					yOff   = imgYW[0]
					yDelta = (imgYW[DimSize(imgYW,0)-1] - imgYW[0])/(DimSize(imgYW,0)-1)
				endif
			elseif(WaveExists(w_x))
				xOff   = w_x[0]
				xDelta = (w_x[DimSize(w_x,0)-1] - w_x[0])/(DimSize(w_x,0)-1)
			endif
			sprintf wOffsets, "%f;%f;", xOff, yOff
			
			ControlInfo/W=$pPath cImageYShift;	int useY = V_Value && isImg
			variable curPos = useY  ? vcsr($csr, s.WinName) : hcsr($csr, s.WinName)
			variable height = isImg ? zcsr($csr, s.WinName) : vcsr($csr, s.WinName)
			
			string offsetWave = GetUserData(pPath, "vRelShift",  "wavname")
			string renormWave = GetUserData(pPath, "vNormalize", "wavname")
			variable startOff = str2num(StringFromList(useY,GetUserData(pPath, "vRelShift", "offset")))
			
			string wUnit = WaveUnits(w,0); 	wUnit[0] = SelectString(strlen(wUnit),""," ")
			TitleBox tShift			,win=$pPath	,title=SelectString(useY,"X pos:","Y pos:")
			TitleBox tRelShift		,win=$pPath	,title=SelectString(useY,"Relative X shift:","Relative Y shift:")
			CheckBox cImageYShift 	,win=$pPath	,disable=!isImg
			CheckBox cImageOneCol	,win=$pPath	,disable=!(WaveDims(w)==2)*2
			TitleBox tCursor		,win=$pPath	,title="Cursor "+csr	,userdata=csr									// the cursor position will be saved into the variable control
			SetVariable vShift		,win=$pPath	,value=_NUM:curPos		,limits={-inf,inf,(useY ? yDelta : xDelta)} ,format="%g"+wUnit
			SetVariable vNormalize	,win=$pPath	,value=_NUM:height		,limits={-inf,inf,height/10}
			if(!StringMatch(wName,renormWave))																			// check if the current and the last waves are the same
				SetVariable vNormalize	,win=$pPath	,userdata(reset)="0" ,userdata(wavname)=wName						// if not, switch over to the new wave
			endif
			if(StringMatch(wName,offsetWave))																			// check if the current and the last waves are the same
				SetVariable vRelShift	,win=$pPath	,value=_NUM:(round(((useY ? yOff : xOff)-startOff)*10000)/10000)	// set control to the updated value
			else
				SetVariable vRelShift	,win=$pPath	,value=_NUM:0 ,limits={-inf,inf,(useY ? yDelta : xDelta)} ,userdata(wavname)=wName ,userdata(offset)=wOffsets ,format="%g "+wUnit
			endif
			if (hv != 0)
				SetVariable vPhoton		,win=$pPath	,value=_NUM:hv ,format="%g "+hUnit
			endif
		break
	EndSwitch
	return HookTakeover
end
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
function QuickScale_PanelHook(STRUCT WMWinHookStruct &s)
	if (s.eventCode == 17)	// kill
		string gName = StringFromList(0, s.WinName, "#")
		HideInfo/W=$gName
		SetWindow $gName hook(MainWindow)=$""													// remove hook from main window
	endif
	return 0
end

//################################################################################################

function QuickRescale_ButtonExecute(STRUCT WMButtonAction &s) : ButtonControl
	if (s.eventCode != 2)
		return 0
	endif
	
	string gName = StringFromList(0, s.win, "#")
	string csr = GetUserData(s.win, "tCursor", "")
	string traceList = "", trace = ""
	variable Findx = 0, Findy = 0
	int items = FetchTraces(gName,traceList)
	
	if (CmpStr(csr,"none") != 0 && strlen(csr))
		wave   yw = CsrWaveRef($csr,gName)
		wave/Z xw = CsrXWaveRef($csr,gName)
		trace = StringByKey("TNAME",CsrInfo($csr,gName))
		Findx = pcsr($csr,gName)
		Findy = qcsr($csr,gName)
	endif
	if (!strlen(trace) && CmpStr(s.ctrlName,"bAddCursor") != 0)
		s.ctrlName = ""
	endif
	
	variable i, GoUp = 0, xOff = 0, yOff = 0, xScale = 0, yScale = 0
	StrSwitch(s.ctrlName)
		case "bFindMax":
			WaveStats/Q yw
			Findx = V_maxRowLoc
			Findy = V_maxColLoc
		break
		case "bFindMin":
			WaveStats/Q yw
			Findx = V_minRowLoc
			Findy = V_minColLoc
		break
		case "bSwitch":
			variable start = DimOffset(yw,0)
			if (WaveExists(xw))
				start = xw[0]
			endif
			ControlInfo/W=$(s.win) vPhoton
			variable hv = V_Value, skip = 0			
			if (numtype(hv) == 0 && hv > 0)
				#if (Exists("QuickRescale_UserFunc") == 6)										// uses user code for photon energy handling
					skip = QuickRescale_UserFunc(yw, xw, hv)
				#endif
				if (skip)
					break
				endif
				//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
				string notes = GetUserData(s.win, "vRelShift", "offset")						// get the initial x offset
				hv = start < 0 ? hv : -hv														// decide the direction
				if ((start+hv)*start <= 0)
					if (WaveExists(xw))
						xw += hv
					else
						SetScale/P x start+hv, DimDelta(yw,0), WaveUnits(yw,0), yw
					endif
					sprintf notes, "%f;%s;",str2num(StringFromList(0,notes))+hv, StringFromList(1,notes)	// write new offset
					SetVariable vRelShift ,win=$(s.win) ,userdata(offset)=notes					// set new offset
				else
					Print "Photon energy is too small."
				endif
			else
				Print "Photon energy needs to be a positive number."
			endif
		break
		case "bNormBack":
			variable oldMax = str2num(GetUserData(s.win, "vNormalize", "reset")), newMax = Wavemax(yw)
			if (oldMax != 0 && numtype(oldMax) == 0)
				yw *= oldMax/newMax																// normalize the wave
				SetVariable vNormalize ,win=$(s.win) ,userdata(reset)="0"
			endif
		break
		case "bAddCursor":
			for (i = 0; i < ItemsInList(kUsedCursorList); i++)									// find first unused cursor
				csr = StringFromList(i,kUsedCursorList)
				if (!strlen(CsrInfo($csr,gName)))
					break
				endif
			endfor
			
			if (!items)																			// if there are no traces, then this must be an image
				trace = StringFromList(0,ImageNameList(gName, ";"))
				wave yw = ImageNameToWaveRef(gName,trace)
			else
				trace = StringFromList(0,traceList)
				wave yw = TraceNameToWaveRef(gName,trace)
				wave/Z xw = XWaveRefFromTrace(gName,trace)										// xw is only needed for traces
			endif
			
			Findx = DimSize(yw,0)/2																// set a new cursor on the center of the first trace or image
			if (!items)
				Findy = DimSize(yw,1)/2
			else
				GetAxis/W=$gName/Q bottom
				if (!V_flag)
					getAllTraceOffsets(gName, trace, xOff, yOff, xScale, yScale)				// center of the current axis corrected for scaling and offset of the current trace
					if (WaveExists(xw))
						FindLevel/P/Q xw, (V_min+(V_max-V_min)/2 - xOff)/xScale
						Findx = (!v_flag && v_levelX > 0 && v_levelX < DimSize(xw,0)-1) ? v_levelX : round(DimSize(xw,0)/2)		// make sure to stay within range
					else
						Findx = x2pnt(yw,(V_min+(V_max-V_min)/2 - xOff)/xScale)
					endif
				endif
			endif
		break
		case "bTraceNext":
			GoUp = 1
		case "bTracePrev":
			if (strlen(ImageInfo(gName,trace,0)))
				break
			endif
			getAllTraceOffsets(gName, trace, xOff, yOff, xScale, yScale)
			Findx = hcsr($csr, gName)*xScale + xOff												// correct for current trace's offsets	
			
			variable curTrace = WhichListItem(trace, traceList)
			if (GoUp)
				if (curTrace < items-1)
					 trace = StringFromList(curTrace+1,traceList)
				endif
			else
				if (curTrace > 0)
					 trace = StringFromList(curTrace-1,traceList)
				endif
			endif
			wave yw = TraceNameToWaveRef(gName, trace)
			wave/Z xw = XWaveRefFromTrace(gName,trace)
			
			getAllTraceOffsets(gName, trace, xOff, yOff, xScale, yScale)						// correct for next trace's offsets
			if (WaveExists(xw))
				FindLevel/P/Q xw, (Findx - xOff)/xScale
				Findx = (!v_flag && v_levelX > 0 && v_levelX < DimSize(xw,0)-1) ? v_levelX : round(DimSize(xw,0)/2)			// make sure to stay within range
			else
				Findx = x2pnt(yw,(Findx - xOff)/xScale)
			endif
		break
	EndSwitch

	if (strlen(ImageInfo(gName,trace,0)))
		Cursor/P/I $csr, $trace, Findx, Findy													// move cursor to image position
	else
		Cursor/P   $csr, $trace, Findx															// move cursor to trace position
	endif
	SetActiveSubwindow $gName																	// restore focus to main window
	return 0
end

//################################################################################################

function QuickRescale_VariableExecute(STRUCT WMSetVariableAction &s) : SetVariableControl
	if (!(s.eventCode == 1 || s.eventCode == 2))
		return 0
	endif
	
	string varStr = s.sval
	string gName = StringFromList(0, s.win, "#")
	string csr = GetUserData(s.win, "tCursor", "")
	if (CmpStr(csr,"none") == 0 || !strlen(csr))
		return 0
	endif
	wave yw = CsrWaveRef($csr, gName)
	wave/Z xw = CsrXWaveRef($csr,gName)
	string trace = StringByKey("TNAME",CsrInfo($csr,gName))
	string imgInfo = ImageInfo(gName,trace,0)
	
	ControlInfo/W=$(s.win) cImageYShift
	variable useY = V_Value && !V_disable, modVal, iniVal
	if (strlen(imgInfo))																		// the xw is used for shifts
		if (useY)
			wave/Z xw = $(StringByKey("YWAVEDF",imgInfo)+StringByKey("YWAVE",imgInfo))
		else
			wave/Z xw = $(StringByKey("XWAVEDF",imgInfo)+StringByKey("XWAVE",imgInfo))
		endif
	endif
	
	StrSwitch(s.ctrlName)
		case "vNormalize":
			modVal = strlen(ImageInfo(gName,trace,0)) ? zcsr($csr,gName)/s.dval : vcsr($csr,gName)/s.dval	// get the 1D y-value or 2D z-value of the cursor
			iniVal = str2num(StringFromList(useY,GetUserData(s.win, "vNormalize", "reset")))
			if (modVal != 0 && numtype(modVal) == 0)
				if (iniVal == 0)
					SetVariable vNormalize ,win=$(s.win) ,userdata(reset)=num2str(Wavemax(yw))				// the original height will be saved into the variable control
				endif
				yw /= modVal
			endif
		break
		case "vShift":
			modVal = useY ? vcsr($csr,gName) : hcsr($csr,gName)
			if (WaveExists(xw))
				xw -= (modVal-s.dval)
				break
			endif
			if (useY)
				SetScale/P y, DimOffset(yw,1)-(modVal-s.dval), DimDelta(yw,1), WaveUnits(yw,1), yw			// scale relative to cursor position in y direction
			else
				SetScale/P x, DimOffset(yw,0)-(modVal-s.dval), DimDelta(yw,0), WaveUnits(yw,0), yw			// scale relative to cursor position in x direction
			endif
		break
		case "vRelShift":
			iniVal = str2num(StringFromList(useY,GetUserData(s.win, "vRelShift", "offset")))
			if (WaveExists(xw))
				modVal = s.dval - (xw[0] - iniVal)												// calculate the difference to the previous change
				xw += modVal
				break
			endif
			if (useY)
				SetScale/P y, iniVal+s.dval, DimDelta(yw,1), WaveUnits(yw,1), yw		
			else
				SetScale/P x, iniVal+s.dval, DimDelta(yw,0), WaveUnits(yw,0), yw
			endif
		break
		case "vArithmetic":
			ControlInfo/W=$(s.win) cImageOneCol;	variable singleCol = V_Value && !V_disable
			DFREF saveDFR = GetDataFolderDFR()
			SetDataFolder GetWavesDataFolderDFR(yw)
			
			string MultiDim = "", columnStr
			if (WaveDims(yw)==2 && singleCol)													// working on 2D waves
				variable col = qcsr($csr,gName)
				if (!strlen(ImageInfo(gName,trace,0)))											// waterfall plot
					SplitString/E=("\]\[(.*)\]") StringByKey("YRANGE",TraceInfo(gName,trace,0)), columnStr		// find the displayed column from TraceInfo
					col = str2num(columnStr)
				endif
				if (numtype(col) == 0)
					MultiDim = "[]["+num2str(col)+"]"
				endif
			endif
			
			string Op = varStr[0]
			if ((CmpStr(Op, "+") == 0) || (CmpStr(Op, "-") == 0) || (CmpStr(Op, "*") == 0) || (CmpStr(Op, "/") == 0))
				string command = PossiblyQuoteName(NameOfWave(yw)) + MultiDim + Op + "=" + varStr[1,inf]			// construct a multiply command string
				Execute/Q/Z command
				if (V_flag)
					Print "There was an error with this command: " + GetErrMessage(V_Flag,2)
				endif
			else
				Print "You need to start your expression by a arithmetic operator +,-,* or /."
			endif
			SetDataFolder saveDFR
			break
	EndSwitch
	
	if (strlen(ImageInfo(gName,trace,0)))														// update the cursor (recalculates the control values)
		Cursor/I/P $csr, $trace, pcsr($csr,gName), pcsr($csr,gName)								// move cursor to image position
	else
		Cursor/P $csr, $trace, pcsr($csr,gName)													// move cursor to trace position
	endif
	SetActiveSubwindow $gName																	// restore focus to main window
	return 0
end

//################################################################################################

function QuickRescale_CheckExecute(STRUCT WMCheckboxAction &s) : CheckBoxControl
	if(s.eventCode == 2)
		string gName = StringFromList(0, s.win, "#")
		string csr = GetUserData(s.win, "tCursor", "")											// extract from sub window
		if (CmpStr(csr,"none") != 0 && strlen(csr))
			string image = CsrWave($csr, gName)													// wave reference from cursor
			if (strlen(ImageInfo(gName,image,0)))												// update the cursor (recalculates the control values)
				Cursor/I/P $csr, $image, pcsr($csr, gName), qcsr($csr, gName)
			endif
		endif
	endif
	return 0
end